# -*- coding: utf-8 -*-
"""
Created on Mon Jun  7 11:39:16 2021

@author:        
    Eva-Maria Grommes

Additional co-authors can be found here:
https://github.com/cire-thk/bifacialSimu    

name:
    BifacialSimu - radiationHandler

overview:
    Manages the calculation of the radiation over the extern modules
    bifacial Radiance and/or PVfactors and delivers the radiation data for the 
    further simulation of BifacialSimu

"""


'Might have to remove IPython functions from this file. IPython should be used in GUI.py only'
from IPython import get_ipython
get_ipython().magic('reset -sf')

import pandas as pd #pandas = can read .csv as input
import matplotlib.pyplot as plt #display shadows
import matplotlib.dates as mdates
import numpy as np
import warnings
import datetime
import numpy
import dateutil.tz
import sys
import math
#import os #to import directories
#from pvlib.location import Location
#from tqdm import tqdm
#import BifacialSimu_dataHandler
#import math
#import pvlib #for electrical output simulation

# DEPENDENCIES AFTER VENDORING
from Vendor.bifacial_radiance.main import RadianceObj, AnalysisObj
from Vendor.pvfactors.viewfactors.aoimethods import faoi_fn_from_pvlib_sandia #to calculate AOI reflection losses
from Vendor.pvfactors.engine import PVEngine
from Vendor.pvfactors import irradiance, geometry, viewfactors

    
class RayTrace:
    
    """
    Raytracing class that uses the bifacial_radiance library to simulate front and rear irradiance of the PVarray.
    Uses parameters passed into BifacialSimuu.main's simulationDict.
 
    Methods
    -------
    createDemo : create a bifacial_radiance RadianceObj that calculates sun parameters and sky values
    simulateRaytrace: Function to perform raytracing simulation. Uses RadianceObj and simulation parameters 
                      for fixed tilt or single axis tracking simulation.  

    """
    
    def createDemo(simulationDict, resultsPath):
        # Create a RadianceObj 'object'
        simulationName = simulationDict['simulationName']
        demo = RadianceObj(name = simulationName, path = resultsPath) # Create a RadianceObj 'object'
        
        return demo

    def simulateRayTrace(simulationDict,demo,metdata, resultsPath, dataFrame, onlyBackscan):
        """
        Function to calculate singleAxisTracking or fixed tilt with the bifacial_radiance library.
        Uses cumulativeSky or gendayLit functions to calculate the irradiance, depending on input parameters in BifacialSimuu_main.py
        Returns df_reportRT for further use in the calculationHandler.
        
        Note: For cumulativeSky, no report will be created, since a  file containing the relevant data is already created by bifacial_radiance.
        Additionally, a cumulativeSky approach is not compatible with Viewfactors calculation, since it averages the sky over 1 year, while VF uses hourly calculations.
        
        Parameters
        ----------
        simulationDict: simulation Dictionary, which can be found in BifacialSimu_main.py
        demo: Bifacial_Radiance's RadianceObj created in "createDemo" function
        metdata: Object containing meteorological data and sun parameters
        resultsPath: output filepath
        dataFrame: helper DataFrame 
        onlyBackscan: option to only calculate rear side of the modules
        """
        
        ####################################################
        
        # Set albedo from sim parameters
        if simulationDict['fixAlbedo'] ==True:
            # Measured Albedo average fix value
            demo.setGround(simulationDict['albedo'])
            
        else:
            # hourly spectral albedo or hourly measured albedo
            if simulationDict['singleAxisTracking'] == True:
                demo.setGround(material = None)
            else:
                sys.exit("The use of hourly Measured Albedo Values is not possible with fixed tilts at the moment")
            
            # if material=None, then material = metdata.albedo 
            # metdata.albedo (datatype: np.ndarray) is colume of metdata = RadianceObj out of weatherfile with albedo values
            # weatherfile must contain hourly measured albedo, if simulationDict['hourlymeasuredAlbedo'] = True
            # weatherfile contains spectral albedo, if simulationDict['hourlyspectralAlbedo'] = True
            # because spectralAlbedoHandler calcutlate spectral albedo and write the calculated spectral albedo in weatherfile
        
        ####################################################    
        
            

        # DEFINE a Module type
        moduley = simulationDict['moduley']*simulationDict['nModsy']
        demo.makeModule(name=simulationDict['module_type'],x=simulationDict['modulex'], y=moduley)
     
        # Make the Scene
        
        sceneDict = {'tilt': simulationDict['tilt'],'gcr': simulationDict['gcr'],'clearance_height':simulationDict['clearance_height'],'hub_height':simulationDict['hub_height'], 'azimuth':simulationDict['azimuth'], 'nModsx': simulationDict['nModsx'], 'nRows': simulationDict['nRows']} 
        
        #################
        # Cumulativ Sky
        
        if simulationDict['cumulativeSky'] == True:
            if simulationDict['singleAxisTracking'] == True:
                # get SingleAxisTracking Data
                trackerdict = demo.set1axis(metdata = metdata, limit_angle = simulationDict['limitAngle'], backtrack = simulationDict['backTracking'], 
                            gcr = simulationDict['gcr'], cumulativesky = True)
               # make the sky
                trackerdict = demo.genCumSky1axis(trackerdict = trackerdict)
                # Make the Scene
                trackerdict = demo.makeScene1axis(trackerdict = trackerdict, moduletype = simulationDict['module_type'], sceneDict = sceneDict) 
                # make oct file
                trackerdict = demo.makeOct1axis(trackerdict = trackerdict)
                results = demo.analysis1axis(trackerdict, onlyBackscan = onlyBackscan)
                #demo.exportTrackerDict(trackerdict = demo.trackerdict, savefile = 'results\\test_reindexTrue.csv', reindex = False)
            else:
                scene = demo.makeScene(simulationDict['module_type'],sceneDict)
                demo.genCumSky(demo.epwfile) # entire year.
                
                octfile = demo.makeOct(demo.getfilelist())  
                analysis = AnalysisObj(octfile, demo.basename)
                frontscan, backscan = analysis.moduleAnalysis(scene)
                results = analysis.analysis(octfile, demo.basename, frontscan, backscan, onlyBackscan = onlyBackscan)  
        
        #################
        # gendayLit
        # Single Axis Tracking
        else:
            if simulationDict['singleAxisTracking'] == True:

                # get SingleAxisTracking Data
                trackerdict = demo.set1axis(metdata = metdata, limit_angle = simulationDict['limitAngle'], backtrack = simulationDict['backTracking'], 
                            gcr = simulationDict['gcr'], cumulativesky = False)
                # make the sky
                startdate = str(simulationDict['startHour'][1]) +'_'+str(simulationDict['startHour'][2])+'_'+str(simulationDict['startHour'][3])
                enddate = str(simulationDict['endHour'][1]) +'_'+str(simulationDict['endHour'][2])+'_'+str(simulationDict['endHour'][3])
                
                trackerdict = demo.gendaylit1axis(startdate=startdate, enddate=enddate) 
                # Make the Scene
                trackerdict = demo.makeScene1axis(trackerdict = trackerdict, moduletype = simulationDict['module_type'], sceneDict = sceneDict)
                
                #demo.makeOct1axis()
                #results_rtrace = demo.analysis1axis(customname="row_" + str(j+1), rowWanted = rowWanted, sensorsy = simulationDict['sensorsy'], onlyBackscan = onlyBackscan) 
                
                
                
                dtStart = datetime.datetime(simulationDict['startHour'][0], simulationDict['startHour'][1], simulationDict['startHour'][2], simulationDict['startHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
                beginning_of_year = datetime.datetime(dtStart.year, 1, 1, tzinfo=dtStart.tzinfo)
                startHour = int((dtStart - beginning_of_year).total_seconds() // 3600)
                
                
                dtEnd = datetime.datetime(simulationDict['endHour'][0], simulationDict['endHour'][1], simulationDict['endHour'][2], simulationDict['endHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
                beginning_of_year = datetime.datetime(dtEnd.year, 1, 1, tzinfo=dtEnd.tzinfo)
                endHour = int((dtEnd - beginning_of_year).total_seconds() // 3600)
                
                ###############
                #Begin raytracing procedure and save data in pd.dataframes
                #df_rtraceFront: Dataframe with raytracing frontdata for all sensors - row-wise
                #df_rtraceBack: Dataframe with raytracing reardata for all sensors - row-wise
                #df_rtrace: All raytracing data for all sensors - row-wise
                #df_reportRT: Dataframe report to be used for electrical simulation, contains mean values for each row
                
                # make oct file
                
                df_reportRT = pd.DataFrame()
                i=0
                demo.makeOct1axis()
                
                for time in range(startHour, endHour+1):
                    
                    x = time - startHour
                    
                    singleindex= dtStart + x*datetime.timedelta(hours=1) 
                    print(singleindex)
                    singleindex = singleindex.strftime('%m_%d_%H')
                    
                    df_rtraceFront = pd.DataFrame()
                    df_rtraceBack = pd.DataFrame()
                    df_rtrace = pd.DataFrame()
                    

                    for j in range(0, simulationDict['nRows']):
        
                        key_front = "row_" + str(j) + "_qinc_front"
                        key_back = "row_" + str(j) + "_qinc_back"
                        
                        rowWanted = j
                        
                        #try if there is data (day) at this time or not (night)
                        try:    
                            results_rtrace = demo.analysis1axis(customname="row_" + str(j), rowWanted = rowWanted, sensorsy = simulationDict['sensorsy'], onlyBackscan = onlyBackscan, singleindex = singleindex) 
                            if onlyBackscan == False:
    
                                df_rtraceFront.insert(loc=j, column = key_front, value = demo.Wm2Front) 
                                df_rtraceBack.insert(loc=j, column = key_back, value = demo.Wm2Back) 
                                df_rtrace = pd.concat([df_rtraceFront, df_rtraceBack], axis=1)
    
                            else:
                                df_rtrace.insert(loc=j, column = key_back, value = demo.Wm2Back) 
                            

                        except:

                        
                            if onlyBackscan == False:
                                df_rtraceFront = pd.DataFrame({key_front: [np.NaN]})
                                df_rtraceBack = pd.DataFrame({key_back: [np.NaN]})
                                df_rtrace_all = pd.concat([df_rtraceFront, df_rtraceBack], axis=1)
                                df_rtrace = df_rtrace.append(df_rtrace_all)
                            else:
                                df_rtraceBack = pd.DataFrame({key_back: [np.NaN]})
                                df_rtrace = df_rtrace.append(df_rtraceBack)
                                

            

                    for j in range(0, simulationDict['nRows']):
                    
                            key_front = "row_" + str(j) + "_qinc_front"
                            key_back = "row_" + str(j) + "_qinc_back"
                            
                            key_front_abs = "row_" + str(j) + "_qabs_front"
                            key_back_abs = "row_" + str(j) + "_qabs_back"
                            
                            
                            if onlyBackscan == False:
                                df_rtrace[key_front] = np.mean(df_rtrace[key_front])       
                                df_rtrace[key_back] = np.mean(df_rtrace[key_back])
                                
                                df_rtrace[key_front_abs] = df_rtrace[key_front] * (1-simulationDict['frontReflect'])   
                                df_rtrace[key_back_abs] = df_rtrace[key_back] * (1-simulationDict['BackReflect'])
                            else:
                                df_rtrace[key_back] = np.mean(df_rtrace[key_back])
                                df_rtrace[key_back_abs] = df_rtrace[key_back] * (1-simulationDict['BackReflect'])
                        
                        
                        
                        
                    df_rtrace = df_rtrace.iloc[:1]        
                    df_reportRT = df_reportRT.append(df_rtrace)
                    
                    i = i+1

                # print(df_rtraceFront)
                # print(df_rtraceBack)
                # print(df_rtrace)
                # print(df_reportRT)
                
                # Set timeindex for report
                    
                df_reportRT.to_csv(resultsPath + "df_reportRT.csv")  
                df_reportRT=df_reportRT.set_index(pd.date_range(start = dtStart - datetime.timedelta(hours=1), end = dtEnd, freq='H', closed='right'))
                
                
                
                #demo.exportTrackerDict(trackerdict = demo.trackerdict, savefile = 'results\\test_reindexTrue.csv', reindex = False)
            
            
            #################
            # gendayLit
            # Fixed tilt
                                   
            else:
                scene = demo.makeScene(simulationDict['module_type'],sceneDict)
                # Translate startHour und endHour in timeindexes
                dtStart = datetime.datetime(simulationDict['startHour'][0], simulationDict['startHour'][1], simulationDict['startHour'][2], simulationDict['startHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
                beginning_of_year = datetime.datetime(dtStart.year, 1, 1, tzinfo=dtStart.tzinfo)
                startHour = int((dtStart - beginning_of_year).total_seconds() // 3600)
                
                dtEnd = datetime.datetime(simulationDict['endHour'][0], simulationDict['endHour'][1], simulationDict['endHour'][2], simulationDict['endHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
                beginning_of_year = datetime.datetime(dtEnd.year, 1, 1, tzinfo=dtEnd.tzinfo)
                endHour = int((dtEnd - beginning_of_year).total_seconds() // 3600)
                
                df = dataFrame
                mask = (df.index >= dtStart) & (df.index <= dtEnd) 
                df_gendaylit= df.loc[mask]
                df_gendaylit = df_gendaylit.reset_index()
                
                
                
                
                
                solpos = metdata.solpos
                #solpos.index[solpos['corrected_timestamp'] == dtStart].tolist()
                
                #mask = (solpos.index >= dtStart) & (solpos.index <= dtEnd)
                
                
                solpos = solpos.loc[mask]
                solpos = solpos.reset_index()
                
                
                df_reportRT = pd.DataFrame()
                i=0
                
                
                
                for time in range(startHour, endHour+1):
                    
                    #dataframes to insert results
                    df_rtraceFront = pd.DataFrame()
                    df_rtraceBack = pd.DataFrame()
                    df_rtrace = pd.DataFrame()
                    
                    # get solar position zenith and azimuth based on site metadata
                    #solpos = pvlib.irradiance.solarposition.get_solarposition(datetimetz,lat,lon,elev)
                    
                    #solpos = solpos.iloc[i]
                    
                    sunalt = float(solpos.loc[i, 'elevation'])
                    sunaz = float(solpos.loc[i, 'azimuth'])-180
                    
                    
                    #sunalt = float(solpos.elevation)
                    # Radiance expects azimuth South = 0, PVlib gives South = 180. Must substract 180 to match.
                    #sunaz = float(solpos.azimuth)-180.0
                    
                    #get dhi and dni out of dataframe
                    #position = time - startHour
                    dni = df_gendaylit.loc[i, 'dni']
                    dhi = df_gendaylit.loc[i, 'dhi']
                    
                    #simulate sky with gendaylit
                    #print("sunalt", sunalt)
                    #print("dhi", dhi)
                    demo.gendaylit2manual(dni, dhi, sunalt, sunaz)
                    #print(time)
                    demo.getfilelist()
                    #print(demo.skyfiles)
                    #print(demo.ground.ReflAvg.shape[0])
                    octfile = demo.makeOct(demo.getfilelist())  

                    
                    analysis = AnalysisObj(octfile, demo.basename)                   
                   
                    for j in range(0, simulationDict['nRows']):
        
                        key_front = "row_" + str(j) + "_qinc_front"
                        key_back = "row_" + str(j) + "_qinc_back"
                        
                        rowWanted = j
                        
                        if octfile != None:
                            
                            frontscan, backscan = analysis.moduleAnalysis(scene, rowWanted=rowWanted, sensorsy=  simulationDict['sensorsy'])
                            results_rtrace = analysis.analysis(octfile, "row_" + str(j), frontscan, backscan, onlyBackscan = onlyBackscan)

                            if onlyBackscan == False:

                                df_rtraceFront.insert(loc=j, column = key_front, value = analysis.Wm2Front) 
                                df_rtraceBack.insert(loc=j, column = key_back, value = analysis.Wm2Back) 
        
                            else:
                                df_rtraceBack.insert(loc=j, column = key_back, value = analysis.Wm2Back) 
                                                           
                            df_rtrace = pd.concat([df_rtraceFront, df_rtraceBack], axis=1)
                            
                        else:
                            if onlyBackscan == False:
                                df_rtraceFront = pd.DataFrame({key_front: [np.NaN]})
                                df_rtraceBack = pd.DataFrame({key_back: [np.NaN]})
                            else:
                                df_rtraceBack = pd.DataFrame({key_back: [np.NaN]})
                                
                            df_rtrace = pd.concat([df_rtraceFront, df_rtraceBack], axis=1)
        
                    if octfile is not None:
                        for j in range(0, simulationDict['nRows']):
                    
                            key_front = "row_" + str(j) + "_qinc_front"
                            key_back = "row_" + str(j) + "_qinc_back"
                            
                            key_front_abs = "row_" + str(j) + "_qabs_front"
                            key_back_abs = "row_" + str(j) + "_qabs_back"
                            
                            
                            if onlyBackscan == False:
                                df_rtrace[key_front] = np.mean(df_rtrace[key_front])       
                                df_rtrace[key_back] = np.mean(df_rtrace[key_back])
                                
                                df_rtrace[key_front_abs] = df_rtrace[key_front] * (1-simulationDict['frontReflect'])   
                                df_rtrace[key_back_abs] = df_rtrace[key_back] * (1-simulationDict['BackReflect'])
                            else:
                                df_rtrace[key_back] = np.mean(df_rtrace[key_back])
                                df_rtrace[key_back_abs] = df_rtrace[key_back] * (1-simulationDict['BackReflect'])
                            
                            
                        df_reportRT = df_reportRT.append(df_rtrace)
                            
                    df_reportRT = df_reportRT.append(df_rtrace)
                    df_reportRT = df_reportRT.iloc[:1+i]
                    i = i+1
                
                
                # Set timeindex for report
                
                df_reportRT=df_reportRT.set_index(pd.date_range(start = dtStart - datetime.timedelta(hours=1), end = dtEnd, freq='H', closed='right'))
                df_reportRT.to_csv(resultsPath + "/df_reportRT.csv")  
                #print(df_rtraceFront)
                #print(df_rtraceBack)
                #print(df_rtrace)
                print(df_reportRT)
                
        return df_reportRT



class ViewFactors:
    
    """
    View class that uses the pvfactors library to simulate front and rear irradiance of the PVarray.
    Uses parameters passed into BifacialSimuu.main's simulationDict.
 
    Methods
    -------
    simulateViewFactors: Function to perform view factor simulation. Uses weather data and sun parameters from bifacial_radiance procedure, which are contained in df.

    """        
    def simulateViewFactors(simulationDict, demo, metdata, dataFrame, resultsPath, onlyFrontscan):
        """
        Function to perform hourly view factor simulation. Can simulate single axis tracking and fixed tilt.
        Can simulate only frontscan or both front and backscan. 
        
        Parameters
        ----------
        simulationDict: simulation Dictionary, which can be found in BifacialSimu_main.py
        demo: Bifacial_Radiance's RadianceObj created in "createDemo" function
        metdata: Object containing meteorological data and sun parameters
        dataFrame: DataFrame containing irradiance data and sun parameters
        resultsPath: output filepath
        onylFrontscan: option to only calculate front side of the modules
        """
        
        
        
        df = dataFrame
        print('view_factor dataframe at beginning of radiation handler:')
        print(df)
        moduley = simulationDict['moduley']*simulationDict['nModsy']
        # Pass simulation parameters over, so that pvfactors doesn't cause errors
        simulationParameter = {
        'n_pvrows': simulationDict['nRows'], #number of PV rows
        'number_of_segments': simulationDict['sensorsy'], #number of segments for each PVrow
        'pvrow_height': simulationDict['clearance_height'], #height of the PV rows, measured at their center [m]
        'pvrow_width': simulationDict['moduley'], #width of the PV panel in row, considered 2D plane [m]
        'pvmodule_width': simulationDict['modulex'], #length of the PV panel in row, considered 2D plane [m]
        'surface_azimuth': simulationDict['azimuth'], #azimuth of the PV surface [deg] 90°= East, 135° = South-East, 180°=South
        'surface_tilt': simulationDict['tilt'], #tilt of the PV surface [deg]
        'albedo': simulationDict['albedo'], # Measured Albedo average value
        #'a0': albedo, # Measured Albedo under direct illumination with a solar zenith angle of approx. 60°
        #'adiff': albedo_diff, # Measured Albedo under 100% diffuse illumination
        'C': 0.4, #Solar angle dependency factor    
        #'index_observed_pvrow': 1, #index of the PV row, whose incident irradiance will be returned
        'rho_front_pvrow' : simulationDict['frontReflect'], #front surface reflectivity of PV rows
        'rho_back_pvrow' : simulationDict['BackReflect'], #back surface reflectivity of PV rows
        'horizon_band_angle' : 6.5, #elevation angle of the sky dome's diffuse horizon band [deg]   
        'L_Position': simulationDict['longitude'], #Longitude of measurement position [deg]
        'L_Area': -105.1727827, #Longitude of timezone area [deg]
        'Latitude_Position': simulationDict['latitude'], #Latitude of measurement position [deg]
        'axis_azimuth': 0.0, #Axis Azimuth angle [deg]
        'gcr': simulationDict['gcr'], #ground coverage ratio (module area / land use)
        'x_min': -100,
        'x_max': 100,
        }
        dpi = 150 #Quality for plot export
        
        
        # Settings for calculating ViewFactor
        get_ipython().run_line_magic('matplotlib', 'qt5')
        np.set_printoptions(precision=3, linewidth=300)
        warnings.filterwarnings('ignore')
        
        # seaborn makes your plots look better
        try:
            import seaborn as sns
            sns.set(rc={"figure.figsize": (12, 6)})
            sns.set_style("darkgrid") # Other available styles: whitegrid, dark, white, ticks
            sns.set_context("paper") # Scaling the axes, also available: talk, poster, 
            sns.set_palette("muted") 
            sns.set_color_codes()
        except ImportError:
            print('We suggest you install seaborn using conda or pip and rerun this cell')
        
        irradiance_model = irradiance.HybridPerezOrdered(rho_front=simulationParameter['rho_front_pvrow'], rho_back=simulationParameter['rho_back_pvrow']) #choose an irradiance model
        # Add dictionary for discretization 
        
        rowSegments = {}
        for i in range(0, simulationParameter['n_pvrows']):
             rowSegments[i] = {'back': simulationParameter['number_of_segments']}
            
        discretization = {'cut':rowSegments}
        simulationParameter.update(discretization)

                        
        #Pick a specific day for closer data consideration
        dayUnderConsideration = 1
        
        df_inputs = df.iloc[dayUnderConsideration * 24:(dayUnderConsideration + 1) * 24, :] #rows to look at in .csv
        
        # Plot the data for displaying direct and diffuse irradiance
        print("\n Direct and Diffuse irradiance:")
        f = plt.Figure(figsize=(12, 3))
        ax1 = f.subplots(1)
        df_inputs[['dni', 'dhi']].plot(ax=ax1)
        ax1.locator_params(tight=True, nbins=6)
        ax1.set_ylabel('W/m2')
        f.savefig("Direct_Diffuse_irradiance.png", dpi = dpi)
        #plt.show()
        #not used to show Plot in own Window
        ##plt.show()(sns)
        
        # Calculate tracking angles if single axis tracking is enabled
        if simulationDict['singleAxisTracking'] == True:
            #create Single Axis Tracking dictionary with bifacialRadiance
            trackerdict = demo.set1axis(metdata = metdata, axis_azimuth = simulationDict['azimuth'],
                                        limit_angle = simulationDict['limitAngle'], 
                                        backtrack = simulationDict['backTracking'],
                                        gcr = simulationDict['gcr'], 
                                        cumulativesky = False)

            # convert trackerdict into dataframe
            d = df.from_dict(trackerdict,
                             orient='index',
                             columns=['dhi','ghi','theta','surf_tilt','surf_azm','ground_clearance'])

            d['time'] = d.index
            # print('tracking dataframe at radiation handler:')
            # print(d)
            d.to_csv(resultsPath + 'd.csv') 
            

            #append variable tilt to data Frame
            df = df.reset_index()
            
            df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
            df = df.set_index('time')
            
            df = df.join(d['surf_tilt'])

            df['timestamp'] = df['corrected_timestamp'].dt.strftime('%m-%d %H:%M%')
            df['timestamp'] = pd.to_datetime(df['timestamp'])
            #df['timestamp'] = df['timestamp'].dt.tz_localize(None)
            df = df.set_index('timestamp')

            surface_tilt = df['surf_tilt']
            print('view_factor dataframe at radiation handler:')
            print(df.index)
            print(df)
        
        else:

            
            df = df.reset_index()
            
            df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
            df = df.set_index('time')
            df['timestamp'] = df['corrected_timestamp'].dt.strftime('%m-%d %H:%M%')
            df['timestamp'] = pd.to_datetime(df['timestamp'])  
            #df['timestamp'] = df['timestamp'].dt.tz_localize(None)
            df = df.set_index('timestamp')
            
            surface_tilt = simulationParameter['surface_tilt']
        
        
        dtStart = datetime.datetime(simulationDict['startHour'][0], simulationDict['startHour'][1], simulationDict['startHour'][2], simulationDict['startHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
        #beginning_of_year = datetime.datetime(dtStart.year, 1, 1, tzinfo=dtStart.tzinfo)
        #startHour = int((dtStart - beginning_of_year).total_seconds() // 3600)
        
        
        dtEnd = datetime.datetime(simulationDict['endHour'][0], simulationDict['endHour'][1], simulationDict['endHour'][2], simulationDict['endHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
        #beginning_of_year = datetime.datetime(dtEnd.year, 1, 1, tzinfo=dtEnd.tzinfo)
        #endHour = int((dtEnd - beginning_of_year).total_seconds() // 3600)



        
        ######### Cutting the dataframe to the required input timeframe
        if simulationDict['cumulativeSky'] == False:
            #df = df.iloc[startHour:endHour]
            mask = (df.index >= dtStart) & (df.index <= dtEnd) 
            df = df.loc[mask]
            
        
        
        ####################################################
        
        # Function to calculate variable albedo according 'PV BIFACIAL YIELD SIMULATION WITH A VARIABLE ALBEDO MODEL' from Matthieu Chiodetti et.al.
        
        def variableAlbedo(df, resultspath):
            
            # Set model parameters
            a0 = 0.22       # Measured Albedo under direct illumination with a solar zenith angle of approx. 60°
            C = 0.4            # Solar angle dependency factor  
            adiff = 0.19896735     # Measured Albedo under 100% diffuse illumination
            
            # Calculate albedo for every hour
            for index, row in df.iterrows(): 
    
                if row['ghi'] == 0: #Avoid division by 0
                    df.loc[index,'albedo'] = 0
                else:
                    df.loc[index,'albedo'] = (1 - row['dhi'] / row['ghi']) * a0 * ((1 + C)  / (1 + 2 * C * math.cos(math.radians(row['zenith'])))) + (row['dhi']  / row['ghi'] * adiff)
      
            if df.loc[index,'albedo'] < 0: #change values below 0 for yield calculation
                df.loc[index,'albedo'] = 0
        
        
            variableAlbedo = pd.DataFrame({'datetime':df.index, 'variable_Albedo': df['albedo']})
            variableAlbedo.to_csv(resultspath + '/variable_Albedo.csv', sep=';', index=False)
            
            # Plot variable albedo
            plt.rc ('axes', labelsize = 13) # Schriftgröße der x- und y-Beschriftungen
            plt.rc ('xtick', labelsize = 11) #Schriftgröße der x-Tick-Labels
            plt.rc ('ytick', labelsize = 11) #Schriftgröße der y-Tick-Labels
            plt.rc ('legend', fontsize = 11) #Schriftgröße der Legende
            f, ax = plt.subplots(figsize=(12, 4), dpi=200)
            
            variableAlbedo[['variable_Albedo']].plot(ax=ax)
            ax.set_ylabel('Variable albedo')
            ax.legend(bbox_to_anchor=(0., 1.02, 1, 0.1), loc='lower left', ncol=2, borderaxespad=0.)
            plt.ylim(0,0.4)
            #plt.show()()
        
        ####################################################
        
        # set Albedo calculation mode
        if simulationDict['fixAlbedo'] == True :
            # Measured Albedo average value, fix value
            albedo = simulationParameter['albedo']
            print("fix", albedo)
       
        elif simulationDict['variableAlbedo'] == True :
            # calculated variable albedo
            variableAlbedo(df, resultsPath)
            albedo = df['albedo']  
            print("variable", albedo)
        else:
            # hourly measured albedo or hourly spectral albedo
            albedo = df['albedo']  
            print("spectral", albedo)
            # weatherfile must contain hourly measured albedo, if simulationDict['hourlymeasuredAlbedo'] = True
            # weatherfile contains spectral albedo, if simulationDict['hourlyspectralAlbedo'] = True
            # because spectralAlbedoHandler calculate spectral albedo and write the calculated spectral albedo in weatherfile
            # weatherfile is read in again in simulationController as df, spectralAlbedo is in df
        ####################################################
        
        #set sun parameters
        
        surface_azimuth = simulationParameter['surface_azimuth']
        
        df['zenith'] =  np.deg2rad(df.zenith)
        df['azimuth'] =  np.deg2rad(df.azimuth)
        
        df['surface_tilt'] = surface_tilt

        df['surface_azimuth'] = surface_azimuth
        
        
        #df = df.rename(columns = {'corrected timestamps': 'time'})
        
        df.to_csv(resultsPath +"/Data.csv")
        ####################################################
        
        # Run full bifacial simulation
        
        # Create ordered PV array and fit engine
        pvarray = geometry.OrderedPVArray.init_from_dict(simulationParameter)
        engine = PVEngine(pvarray)

        engine.fit(df.index, df.dni, df.dhi,
                   df.azimuth, df.zenith,
                   df.surface_tilt, df.surface_azimuth,
                   albedo)
        
              
        
        # Devide PV array into segments
        number_of_segments = {} # create empty list
        
        # Create loop for defining the number of segments
        for i in range(0, simulationParameter['n_pvrows']):
            number_of_segments[i] = len(pvarray.ts_pvrows[i].back.list_segments)
        """
        # Print Plot including Segments
        #print("\n Segment Division:")    
        f, ax = plt.subplots(figsize = (12,4))
        pvarray.plot_at_idx(1,ax,with_surface_index = True)
        ax.set_xlim(-3,20)
        #f.savefig(resultsPath +"/Segment_Division" + datetime.now().strftime("%Y-%m-%d-%H-%M") + ".png", dpi = dpi)
        f.savefig("Segment_Division.png", dpi = dpi)
        #plt.show()(sns)
        """
        ####################################################
        #AOI reflection losses
        
        module_name = 'SunPower_128_Cell_Module___2009_'
        
        # Create an faoi function
        faoi_function = faoi_fn_from_pvlib_sandia(module_name)
        
        
        # Helper functions for plotting and simulation with reflection losses
        def plot_irradiance(df_reportVF):
        
            # Plot irradiance
            f = plt.Figure(figsize=(12, 4))
            ax = f.subplots(nrows=1, ncols=2)
            
            # Plot back surface irradiance
            df_reportVF[['qinc_back', 'qabs_back']].plot(ax=ax[0])
            ax[0].set_title('Back surface irradiance')
            ax[0].set_ylabel('W/m2')
            
            # Plot front surface irradiance
            df_reportVF[['qinc_front', 'qabs_front']].plot(ax=ax[1])
            ax[1].set_title('Front surface irradiance')
            ax[1].set_ylabel('W/m2')
            #plt.show()()
            
            
        ##### Without backscan #####   
        def plot_irradiance_front(df_reportVF):
            
                
            # Plot irradiance
            f = plt.Figure(figsize=(12, 4))
            ax = f.subplots()
                    
            # Plot front surface irradiance
            df_reportVF[['qinc_front', 'qabs_front']].plot(ax=ax)
            ax.set_title('Front surface irradiance')
            ax.set_ylabel('W/m2')
            #plt.show()()
        
        
        def plot_aoi_losses(df_reportVF):
                # plotting AOI losses
                f = plt.Figure(figsize=(5.5, 4))
                ax = f.subplots()
                df_reportVF[['aoi_losses_back_%']].plot(ax=ax)
                df_reportVF[['aoi_losses_front_%']].plot(ax=ax)
                
                # Adjust axes
                ax.set_ylabel('%')
                ax.legend(['AOI losses back PV row', 'AOI losses front PV row'])
                ax.set_title('AOI losses')
                #plt.show()()
        
        
        ##### Without backscan #####
        def plot_aoi_losses_front(df_reportVF):
            
            # plotting AOI losses
            f = plt.Figure(figsize=(5.5, 4))
            ax = f.subplots()
            df_reportVF[['aoi_losses_front_%']].plot(ax=ax)
            
            # Adjust axes
            ax.set_ylabel('%')
            ax.legend(['AOI losses front PV row'])
            ax.set_title('AOI losses')
            #plt.show()()
            
           
            
        # Create a function that will build a simulation report      
        def fn_report(pvarray):
            if simulationDict['nRows']==1:
                n = 0
            else:
                n = 1        
            reportAOI = {'qinc_back': pvarray.ts_pvrows[n].back.get_param_weighted('qinc'),
                      'qabs_back': pvarray.ts_pvrows[n].back.get_param_weighted('qabs'),
                      'qinc_front': pvarray.ts_pvrows[n].front.get_param_weighted('qinc'),
                      'qabs_front': pvarray.ts_pvrows[n].front.get_param_weighted('qabs')}
            
            # Calculate AOI losses
            reportAOI['aoi_losses_back_%'] = (reportAOI['qinc_back'] - reportAOI['qabs_back']) / reportAOI['qinc_back'] * 100.
            reportAOI['aoi_losses_front_%'] = (reportAOI['qinc_front'] - reportAOI['qabs_front']) / reportAOI['qinc_front'] * 100.
            # Return report
            return reportAOI
            
        ##### Without backscan #####
        def fn_report_front(pvarray):
            # Get irradiance values
            if simulationDict['nRows']==1:
                n = 0
            else:
                n = 1
            reportAOI = {
                      'qinc_front': pvarray.ts_pvrows[n].front.get_param_weighted('qinc'),
                      'qabs_front': pvarray.ts_pvrows[n].front.get_param_weighted('qabs')}
            
            # Calculate AOI losses
            
            reportAOI['aoi_losses_front_%'] = (reportAOI['qinc_front'] - reportAOI['qabs_front']) / reportAOI['qinc_front'] * 100.
            
            
            # Return report
            return reportAOI
        
        # Run full mode simulation
            
        if onlyFrontscan == False:
            reportAOI = engine.run_full_mode(fn_build_report=fn_report)
            df_report_AOI = pd.DataFrame(reportAOI, index=df.index)
            plot_irradiance(df_report_AOI)
            
            
        else:
            reportAOI = engine.run_full_mode(fn_build_report=fn_report_front)
            df_report_AOI = pd.DataFrame(reportAOI, index=df.index)
            plot_irradiance_front(df_report_AOI)
        
        
        
        
        ####################################################
        # Define Function calculating the total incident irradiance for the front and back side and the different segments of the backside
        
        # qinc = total incident irradiance on a surface, and it does not account for reflection losses [W/m2]
        # qabs = total absorbed irradiance by a surface [W/m2]
        
        #ts=timeseries 
        
        
              
        def Segments_report(pvarray):
            result = dict()
            
            rowsum_qabs_front = 0
            rowsum_qabs_back = 0
            rowsum_qinc_front = 0
            rowsum_qinc_back = 0
            
            for i in range(0, len(pvarray.ts_pvrows)):
                
                row = pvarray.ts_pvrows[i]
                
                #result["row_" + str(i) + "_reflection_front"] = row.front.get_param_weighted('reflection') #avg reflection for every row front
                #result["row_" + str(i) + "_reflection_back"] = row.back.get_param_weighted('reflection') #avg reflection for every row back
                result["row_" + str(i) + "_qinc_front"] = row.front.get_param_weighted('qinc') #avg qinc for every row front
                result["row_" + str(i) + "_qinc_back"] = row.back.get_param_weighted('qinc') #avg qinc for every row back
                result["row_" + str(i) + "_qabs_front"] = row.front.get_param_weighted('qabs') #avg qabs for every row front
                result["row_" + str(i) + "_qabs_back"] = row.back.get_param_weighted('qabs') #avg qabs for every row back
                #result["row_" + str(i) + "_q0_front"] = row.front.get_param_weighted('q0') #avg reflection for every row front
                #result["row_" + str(i) + "_q0_back"] = row.back.get_param_weighted('q0') #avg reflection for every row back
                #result["row_" + str(i) + "_isotropic_front"] = row.front.get_param_weighted('isotropic') #avg reflection for every row front
                #result["row_" + str(i) + "_isotropic_back"] = row.back.get_param_weighted('isotropic') #avg reflection for every row back              
                
                for ts_surface in row.front.all_ts_surfaces:
                    key = "qabs_segment_" + str(ts_surface.index)  # updated
                    result[key] = ts_surface.get_param('qabs')
                 
                for ts_surface in row.back.all_ts_surfaces:
                    key = "qabs_segment_" + str(ts_surface.index)  # updated
                    result[key] = ts_surface.get_param('qabs')
                    
                for ts_surface in row.front.all_ts_surfaces:
                    key = "qinc_segment_" + str(ts_surface.index)  # updated
                    result[key] = ts_surface.get_param('qinc')
                  
                for ts_surface in row.back.all_ts_surfaces:
                    key = "qinc_segment_" + str(ts_surface.index)  # udpated
                    result[key] = ts_surface.get_param('qinc')
                
                rowsum_qabs_front += row.front.get_param_weighted('qabs')
                rowsum_qabs_back += row.back.get_param_weighted('qabs')
                rowsum_qinc_front += row.front.get_param_weighted('qinc')
                rowsum_qinc_back += row.back.get_param_weighted('qinc')
                
            result["row_avg_qabs_front"] = rowsum_qabs_front / simulationDict['nRows']
            result["row_avg_qabs_back"] = rowsum_qabs_back / simulationDict['nRows']
            result["row_avg_qinc_front"] = rowsum_qinc_front / simulationDict['nRows']
            result["row_avg_qinc_back"] = rowsum_qinc_back / simulationDict['nRows']
            
            return result
               
        def Segments_report_front(pvarray):
            result = dict()
            
            for i in range(0, len(pvarray.ts_pvrows)):
                
                row = pvarray.ts_pvrows[i]
                
                result["row_" + str(i) + "_qabs_front"] = row.front.get_param_weighted('qabs') #avg qabs for every row front
                result["row_" + str(i) + "_qinc_front"] = row.front.get_param_weighted('qinc') #avg qinc for every row front
                
                
                for ts_surface in row.front.all_ts_surfaces:
                    key = "qabs_segment_" + str(ts_surface.index)  # updated
                    result[key] = ts_surface.get_param('qabs')
                 
                            
                for ts_surface in row.front.all_ts_surfaces:
                    key = "qinc_segment_" + str(ts_surface.index)  # updated
                    result[key] = ts_surface.get_param('qinc')
                  
                                   
        # Display length of every segment
                #for ts_surface in row.front.all_ts_surfaces:
                 #   key = "segment_" + str(ts_surface.index) + "length" # all shaded timeseries surfaces on the front side of the PV row have length zero.
                  #  result[key] = ts_surface.length.tolist()
                    
                #for ts_surface in row.back.all_ts_surfaces:
                 #   key = "segment_" + str(ts_surface.index) + "length"
                  #  result[key] = ts_surface.length.tolist()
                    
        # Display if segments are shaded or not            
        #        for ts_surface in row.front.all_ts_surfaces:
        #            key = "segment_" + str(ts_surface.index) + "shaded"
        #            result[key] = ts_surface.shaded.tolist()
                    
        #        for ts_surface in row.back.all_ts_surfaces:
        #            key = "segment_" + str(ts_surface.index) + "shaded"
        #            result[key] = ts_surface.shaded.tolist()    
            
            return result
        
        ####################################################
        
        # function to average the hourly front and back irradiances of the pv rows to daily values
        # only possible, when the starthour and endhour is 0 o'clock, so whole days are simulated
        # otherwise, the loop has to be programmed, that it checks, if there are less than 24 hour for a day and reduces the loops the the acutal hours in the first or last day
        
        def daily_mean_irradiance(df_reportVF):
            df1 = df_reportVF  

            Avg_front_daily = []     # array to hold daily row average qabs front values
            Avg_back_daily = []      # array to hold daily row average qabs back values
            Row_0_front_daily = []   # array to hold daily qabs front values of row 0
            Row_1_front_daily = []   # array to hold daily qabs front values of row 1
            Row_2_front_daily = []   # array to hold daily qabs front values of row 2
            Row_0_back_daily = []   # array to hold daily qabs back values of row 0
            Row_1_back_daily = []   # array to hold daily qabs back values of row 1
            Row_2_back_daily = []   # array to hold daily qabs back values of row 2
            cd = []             # array to hold hourly datetime

     
            dtStart = datetime.date(simulationDict['startHour'][0], simulationDict['startHour'][1], simulationDict['startHour'][2])
            dtEnd = datetime.date(simulationDict['endHour'][0], simulationDict['endHour'][1], simulationDict['endHour'][2])

            delta_days= dtEnd - dtStart
            #print("delta days", delta_days.days)

            for i in range(delta_days.days):
                
                Avg_front_hourly = []     # array to hold hourly row average qabs front values
                Avg_back_hourly = []     # array to hold hourly row average qabs front values
                Row_0_front_hourly = []   # array to hold hourly qabs front values of row 0
                Row_1_front_hourly = []   # array to hold hourly qabs front values of row 1
                Row_2_front_hourly = []   # array to hold hourly qabs front values of row 2
                Row_0_back_hourly = []   # array to hold hourly qabs back values of row 0
                Row_1_back_hourly = []   # array to hold hourly qabs back values of row 1
                Row_2_back_hourly = []   # array to hold hourly qabs back values of row 2
                
                currentDate = datetime.datetime(simulationDict['startHour'][0], simulationDict['startHour'][1], simulationDict['startHour'][2], simulationDict['startHour'][3]) + pd.to_timedelta(i, unit='D') 
                
                for j in range(24):
                    
                    k = i*24 + j
                    
                    Avg_front = df1.iloc[k]['row_avg_qabs_front']    
                    Avg_back = df1.iloc[k]['row_avg_qabs_back'] 
                    Row_0_front = df1.iloc[k]['row_0_qabs_front'] 
                    Row_1_front = df1.iloc[k]['row_1_qabs_front'] 
                    Row_2_front = df1.iloc[k]['row_2_qabs_front'] 
                    Row_0_back = df1.iloc[k]['row_0_qabs_back'] 
                    Row_1_back = df1.iloc[k]['row_1_qabs_back'] 
                    Row_2_back = df1.iloc[k]['row_2_qabs_back'] 
                                                            
                    if Avg_front == 0:
                        Avg_front = np.nan
                    if Avg_back == 0:
                        Avg_back = np.nan
                    if Row_0_front == 0:
                        Row_0_front = np.nan
                    if Row_1_front == 0:
                        Row_1_front = np.nan
                    if Row_2_front == 0:
                        Row_2_front = np.nan
                    if Row_0_back == 0:
                        Row_0_back = np.nan
                    if Row_1_back == 0:
                        Row_1_back = np.nan
                    if Row_2_back == 0:
                        Row_2_back = np.nan
                        
                    Avg_front_hourly.append(Avg_front)
                    Avg_back_hourly.append(Avg_back)
                    Row_0_front_hourly.append(Row_0_front)
                    Row_1_front_hourly.append(Row_1_front)
                    Row_2_front_hourly.append(Row_2_front)
                    Row_0_back_hourly.append(Row_0_back)
                    Row_1_back_hourly.append(Row_1_back)
                    Row_2_back_hourly.append(Row_2_back)
                    

                Avg_front_mean = np.nanmean(Avg_front_hourly)      # mean value of 24 hourly values without nan values
                Avg_back_mean = np.nanmean(Avg_back_hourly)        # mean value of 24 hourly values without nan values
                Row_0_front_mean = np.nanmean(Row_0_front_hourly)  # mean value of 24 hourly values without nan values         
                Row_1_front_mean = np.nanmean(Row_1_front_hourly)  # mean value of 24 hourly values without nan values         
                Row_2_front_mean = np.nanmean(Row_2_front_hourly)  # mean value of 24 hourly values without nan values         
                Row_0_back_mean = np.nanmean(Row_0_back_hourly)    # mean value of 24 hourly values without nan values         
                Row_1_back_mean = np.nanmean(Row_1_back_hourly)    # mean value of 24 hourly values without nan values         
                Row_2_back_mean = np.nanmean(Row_2_back_hourly)    # mean value of 24 hourly values without nan values         
                
                Avg_front_daily.append(Avg_front_mean)
                Avg_back_daily.append(Avg_back_mean)
                Row_0_front_daily.append(Row_0_front_mean)
                Row_1_front_daily.append(Row_1_front_mean)
                Row_2_front_daily.append(Row_2_front_mean) 
                Row_0_back_daily.append(Row_0_back_mean)
                Row_1_back_daily.append(Row_1_back_mean)
                Row_2_back_daily.append(Row_2_back_mean)
                
                cd.append(currentDate)          # append the currentDate to cd array
            
            # create pandas dataframe to save the four arrays and give them headers
            df2 = pd.DataFrame({'datetime':cd, 'Average front surface irradiance': Avg_front_daily, 'Average rear surface irradiance': Avg_back_daily, 'Front surface irradiance row 1':Row_0_front_daily, 'Front surface irradiance row 2':Row_1_front_daily,'Front surface irradiance row 3':Row_2_front_daily,'Rear surface irradiance row 1':Row_0_back_daily, 'Rear surface irradiance row 2':Row_1_back_daily, 'Rear surface irradiance row 3':Row_2_back_daily})
            df2.set_index('datetime')
            print(df2)
            return df2
        
        def plot_irradiance1(df2):
            # Plot average surface irradiance (qabs)
            
            plt.rc ('axes', labelsize = 13) # Schriftgröße der x- und y-Beschriftungen
            plt.rc ('xtick', labelsize = 11) #Schriftgröße der x-Tick-Labels
            plt.rc ('ytick', labelsize = 11) #Schriftgröße der y-Tick-Labels
            plt.rc ('legend', fontsize = 11) #Schriftgröße der Legende
            fig = plt.Figure(figsize=(12, 4), dpi=200)
            ax = fig.subplots()
            width = 1
                       
            y1 = df2['Average front surface irradiance']
            y2 = df2['Average rear surface irradiance']
            ind = range(len(y1))
            ax.bar(ind, y1, width, label='Average front surface irradiance', linewidth=0)
            ax.bar(ind, y2, width, bottom=y1, label='Average rear surface irradiance', linewidth=0)
            
            ax.set_ylabel('Average surface irradiance in $\mathregular{W/m^2}$')
            ax.xaxis.set_major_locator(mdates.MonthLocator())
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%b'))
            ax.legend(bbox_to_anchor=(0., 1.02, 1, 0.1), loc='lower left', ncol=2, borderaxespad=0.)
            #plt.xlim(0,len(y1)-1)
            
            
            #plt.show()()
            
            
        def plot_irradiance2(df2):
            # Plot surface irradiance for every row
            fig = plt.Figure(figsize=(12, 4), dpi=200)
            ax = fig.subplots()
            
            width = 1
                       
            y1 = df2['Rear surface irradiance row 1']
            y2 = df2['Rear surface irradiance row 2']
            y3 = df2['Rear surface irradiance row 3']
            ind = range(len(y1))
            ax.bar(ind, y1, width, label='Rear surface irradiance row 1', linewidth=0)
            ax.bar(ind, y2, width, label='Rear surface irradiance row 2', linewidth=0)
            ax.bar(ind, y3, width, label='Rear surface irradiance row 3', linewidth=0)
            
            ax.xaxis.set_major_locator(mdates.MonthLocator())
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%b'))
            ax.legend(bbox_to_anchor=(0., 1.02, 1, 0.1), loc='lower left', ncol=3, borderaxespad=0.)
            ax.set_ylabel('Rear surface irradiance in $\mathregular{W/m^2}$')
            #plt.xlim(0,len(y1)-1)
            #plt.show()()
        
        
        # Run full simulation
            
        if onlyFrontscan == False:
            report = engine.run_full_mode(fn_build_report=Segments_report)
            df_reportVF = pd.DataFrame(report, index=df.index)
            df2 = daily_mean_irradiance(df_reportVF)  # erzeugt dataframe mit gemittelten täglichen irradiances
            plot_irradiance1(df2)    # Plot mit der durschnittlichen front und back irradiance aller Reihen für jeden Tag
            plot_irradiance2(df2)    # Plot mit der front und back irradiance für jede Reihen für jeden Tag
            
            # Print results as .csv in directory
            df_reportVF.to_csv(resultsPath + "radiation_qabs_results.csv")
            """
            # Plot total qinc front and back for every row
            f, ax = plt.subplots(3, figsize=(10, 6))
            ax.locator_params(tight=True, nbins=5)
            df_reportVF[['row_0_qinc_front', 'row_0_qinc_back']].plot(ax=ax[0])
            df_reportVF[['row_1_qinc_front', 'row_1_qinc_back']].plot(ax=ax[1])
            df_reportVF[['row_2_qinc_front', 'row_2_qinc_back']].plot(ax=ax[2])
            ax[0].set_ylabel('W/m2')
            ax[1].set_ylabel('W/m2')
            ax[2].set_ylabel('W/m2')
            #f.savefig("row0-3_qinc.png", dpi = dpi)
            #plt.show()(sns)
            """
            
            
            
            
        else:
            report = engine.run_full_mode(fn_build_report=Segments_report_front)
            
            # Print results as .csv in directory
            df_reportVF = pd.DataFrame(report, index=df.index)
            df_reportVF.to_csv("radiation_qabs_results.csv")
            
            """
            # Plot total qinc front and back for every row
            f, ax = plt.subplots(3, figsize=(10, 6))
            ax.locator_params(tight=True, nbins=5)
            df_reportVF[['row_0_qinc_front']].plot(ax=ax[0])
            df_reportVF[['row_1_qinc_front']].plot(ax=ax[1])
            df_reportVF[['row_2_qinc_front']].plot(ax=ax[2])
            ax[0].set_ylabel('W/m2')
            ax[1].set_ylabel('W/m2')
            ax[2].set_ylabel('W/m2')
            f.savefig("row0-3_qinc.png", dpi = dpi)
            #plt.show()(sns)
            """
        
        df_reportVF=df_reportVF.set_index(pd.date_range(start = dtStart - datetime.timedelta(hours=1), end = dtEnd, freq='H', closed='right'))
        
        print("df_reportVF at end of RadiationHandler: ")
        print(df_reportVF)

            
        
            

        #df_reportVF.to_csv(resultsPath + "/radiation_qabs_results_" + datetime.now().strftime("%Y-%m-%d-%H-%M") + ".csv")
        
        #f.savefig(resultsPath +"/row0-3_qinc" + datetime.now().strftime("%Y-%m-%d-%H-%M") + ".png", dpi = dpi)
                
        ####################################################
        # Calculate ViewFactors for Day of Consideration
        
        # Instantiate calculator
        vf_calculator = viewfactors.VFCalculator()
        
        # Calculate view factor matrix of the pv array
        vf_matrix = vf_calculator.build_ts_vf_matrix(pvarray)
               
        # Create ViewFactor matrix 
        def save_view_factor(i, j, vf_matrix, timestamps):
            
            vf = vf_matrix[i, j, :]
            matrix = np.around(vf, decimals=2)
            
            result = pd.DataFrame()
            
            # Print table with timestamps and ViewFactors
            result["timestamps"] = timestamps
            result["view_factors"] = matrix.tolist()
            
            result = result.set_index('timestamps')
            
            #result.to_csv(resultsPath + "/view_factors_" + str(i) + "_" + str(j) + "_" + datetime.now().strftime("%Y-%m-%d-%H-%M") + ".csv") # Print ViewFactors to directory
            result.to_csv("view_factors_" + str(i) + "_" + str(j) + ".csv")
            #print("\n View Factors:")
            #print('View factor from surface {} to surface {}: {}'.format(i, j, np.around(vf, decimals=2))) # in case the matrix should be printed in the console
        
        save_view_factor(4, 12, vf_matrix, df.index)
        
        
        f = plt.Figure(figsize=(10, 3))
        ax = f.subplots()
        pvarray.plot_at_idx(0, ax, with_surface_index=True)
        #plt.show()()
        
        return df_reportVF, df


# -*- coding: utf-8 -*-
"""
@author: cdeline

ModuleObj class for defining module geometry

"""
import os
import numpy as np

from bifacial_radiance.main import  _popen, DATA_PATH
 
class SuperClass:
    def __repr__(self):
        return str(self.getDataDict())
    def getDataDict(self):
        """
        return dictionary values from self.  Originally stored as self.data
        """
        return dict(zip(self.keys,[getattr(self,k) for k in self.keys]))   

class ModuleObj(SuperClass):
    """
    Module object to store module & torque tube details.  
    Does the heavy lifting of demo.makeModule()
    Module details are passed in and stored in module.json.
    Pass this object into makeScene or makeScene1axis.
    """

    
    def __init__(self, name=None, x=None, y=None, z=None, bifi=1, modulefile=None, 
                 text=None, customtext='', xgap=0.01, ygap=0.0, zgap=0.1,
                 numpanels=1, rewriteModulefile=True, cellModule=None,  
                 glass=False, modulematerial='black', tubeParams=None,
                 frameParams=None, omegaParams=None, hpc=False):
        """
        Add module details to the .JSON module config file module.json

        Module definitions assume that the module .rad file is defined
        with zero tilt, centered along the x-axis and y-axis for the center
        of rotation of the module (+X/2, -X/2, +Y/2, -Y/2 on each side).
        Tip: to define a module that is in 'portrait' mode, y > x. 

        Parameters
        ------------
        name : str
            Input to name the module type
        x : numeric
            Width of module along the axis of the torque tube or rack. (meters)
        y : numeric
            Length of module (meters)
        bifi : numeric
            Bifaciality of the panel (not currently used). Between 0 (monofacial) 
            and 1, default 1.
        modulefile : str
            Existing radfile location in \objects.  Otherwise a default value is used
        text : str
            Text used in the radfile to generate the module
        customtext : str
            Added-text used in the radfile to generate any
            extra details in the racking/module. Does not overwrite
            generated module (unlike "text"), but adds to it at the end.
        rewriteModulefile : bool
            Default True. Will rewrite module file each time makeModule is run.
        numpanels : int
            Number of modules arrayed in the Y-direction. e.g.
            1-up or 2-up, etc. (supports any number for carport/Mesa simulations)
        xgap : float
            Panel space in X direction. Separation between modules in a row.
        ygap : float
            Gap between modules arrayed in the Y-direction if any.
        zgap : float
            Distance behind the modules in the z-direction to the edge of the tube (m)
        glass : bool
            Add 5mm front and back glass to the module (glass/glass). Warning:
                glass increases the analysis variability.  Recommend setting
                accuracy='high' in AnalysisObj.analysis()
        cellModule : dict
            Dictionary with input parameters for creating a cell-level module.
            Shortcut for ModuleObj.addCellModule()
        tubeParams : dict
            Dictionary with input parameters for creating a torque tube as part of the 
            module. Shortcut for ModuleObj.addTorquetube()
        frameParams : dict
            Dictionary with input parameters for creating a frame as part of the module.
            Shortcut for ModuleObj.addFrame()
        omegaParams : dict
            Dictionary with input parameters for creating a omega or module support 
            structure. Shortcut for ModuleObj.addOmega()
        hpc : bool (default False)
            Set up module in HPC mode.  Namely turn off read/write to module.json
            and just pass along the details in the module object. Note that 
            calling e.g. addTorquetube() after this will tend to write to the
            module.json so pass all geometry parameters at once in to makeModule
            for best response.
        
        
        '"""
        self.keys = ['x', 'y', 'z', 'modulematerial', 'scenex','sceney',
            'scenez','numpanels','bifi','text','modulefile', 'glass',
            'offsetfromaxis','xgap','ygap','zgap' ] 
        
        #replace whitespace with underlines. what about \n and other weird characters?
        # TODO: Address above comment?        
        self.name = str(name).strip().replace(' ', '_') 
        self.customtext = customtext
        
        # are we writing to JSON with passed data or just reading existing?
        if (x is None) & (y is None) & (cellModule is None):
            #just read in file. If .rad file doesn't exist, make it.
            self.readModule(name=name)
            if name is not None:
                self._saveModule(savedata=None, json=False, 
                                 rewriteModulefile=False)

        else:
            # set initial variables that aren't passed in 
            scenex = sceney = scenez = offsetfromaxis = 0
            """
            # TODO: this is kind of confusing and should probably be changed
            # set torque tube internal dictionary
            tubeBool = torquetube
            torquetube = {'bool':tubeBool,
                          'diameter':tubeParams['diameter'],
                          'tubetype':tubeParams['tubetype'],
                          'material':tubeParams['material']
                          }   
            try:
                self.axisofrotationTorqueTube = tubeParams['axisofrotation']
            except AttributeError:
                self.axisofrotationTorqueTube = False
            """
            if tubeParams:
                if 'bool' in tubeParams:  # backward compatible with pre-0.4
                    tubeParams['visible'] = tubeParams.pop('bool')
                if 'torqueTubeMaterial' in tubeParams:  #  pre-0.4
                    tubeParams['material'] = tubeParams.pop('torqueTubeMaterial')
                self.addTorquetube(**tubeParams, recompile=False)
            if omegaParams:
                self.addOmega(**omegaParams, recompile=False)
                
            if frameParams:
                self.addFrame(**frameParams, recompile=False)
                
            if cellModule:
                self.addCellModule(**cellModule, recompile=False)
            
            # set data object attributes from datakey list. 
            for key in self.keys:
                setattr(self, key, eval(key))      
            
            if self.modulefile is None:
                self.modulefile = os.path.join('objects',
                                                       self.name + '.rad')
                print("\nModule Name:", self.name)
                  
            if hpc:
                self.compileText(rewriteModulefile, json=False)
            else:
                self.compileText(rewriteModulefile)
            
    def compileText(self, rewriteModulefile=True, json=True):
        """
        Generate the text for the module .rad file based on ModuleObj attributes.
        Optionally save details to the module.json and module.rad files.

        Parameters
        ------------
        rewriteModulefile : bool (default True)
            Overwrite the .rad file for the module
        json : bool  (default True)
            Update the module.json file with ModuleObj attributes
        """
        saveDict = self.getDataDict()

        if hasattr(self,'cellModule'):
            saveDict = {**saveDict, 'cellModule':self.cellModule.getDataDict()}
        if hasattr(self,'torquetube'):
            saveDict = {**saveDict, 'torquetube':self.torquetube.getDataDict()}
        if hasattr(self,'omega'):
            saveDict = {**saveDict, 'omegaParams':self.omega.getDataDict()}
        if hasattr(self,'frame'):
            saveDict = {**saveDict, 'frameParams':self.frame.getDataDict()}            
        self._makeModuleFromDict(**saveDict)  

        #write JSON data out and write radfile if it doesn't exist
        self._saveModule({**saveDict, **self.getDataDict()}, json=json, 
                         rewriteModulefile=rewriteModulefile)

            
    def readModule(self, name=None):
        """
        Read in available modules in module.json.  If a specific module name is
        passed, return those details into the SceneObj. Otherwise 
        return available module list.

        Parameters:  name (str)  Name of module to be read

        Returns:  moduleDict dictionary or list of modulenames if name is not passed in.

        """
        import json
        filedir = os.path.join(DATA_PATH,'module.json')
        with open( filedir ) as configfile:
            data = json.load(configfile)

        modulenames = data.keys()
        if name is None:
            return list(modulenames)

        if name in modulenames:
            moduleDict = data[name]
            self.name = name
            # BACKWARDS COMPATIBILITY - look for missing keys
            if not 'scenex' in moduleDict:
                moduleDict['scenex'] = moduleDict['x']
            if not 'sceney' in moduleDict:
                moduleDict['sceney'] = moduleDict['y']
            if not 'offsetfromaxis' in moduleDict:
                moduleDict['offsetfromaxis'] = 0
            if not 'modulematerial' in moduleDict:
                moduleDict['modulematerial'] = 'black'
            if not 'glass' in moduleDict:
                moduleDict['glass'] = False    
            if not 'z' in moduleDict:
                moduleDict['z'] = 0.02
            # set ModuleObj attributes from moduleDict
            #self.data = moduleDict
            for keys in moduleDict:
                setattr(self, keys, moduleDict[keys])
            
            # Run torquetube, frame, omega, cellmodule
            if moduleDict.get('torquetube'):
                tubeParams = moduleDict['torquetube']
                if 'bool' in tubeParams:  # backward compatible with pre-0.4
                    tubeParams['visible'] = tubeParams.pop('bool')
                if 'torqueTubeMaterial' in tubeParams:  #  pre-0.4
                    tubeParams['material'] = tubeParams.pop('torqueTubeMaterial')
                self.addTorquetube(**tubeParams, recompile=False)
            if moduleDict.get('cellModule'):
                self.addCellModule(**moduleDict['cellModule'], recompile=False)
            if moduleDict.get('omegaParams'):
                self.addOmega(**moduleDict['omegaParams'], recompile=False) 
            if moduleDict.get('frameParams'):
                self.addFrame(**moduleDict['frameParams'], recompile=False) 
            
            
            return moduleDict
        else:
            print('Error: module name {} doesnt exist'.format(name))
            return {}


    def _saveModule(self, savedata, json=True, rewriteModulefile=True):
        """
        write out changes to module.json and make radfile if it doesn't
        exist.  if rewriteModulefile is true, always overwrite Radfile.

        Parameters
        ----------
        json : bool, default is True.  Save JSON
        rewriteModulefile : bool, default is True.

        """
        import json as jsonmodule
        
        if json:
            filedir = os.path.join(DATA_PATH, 'module.json') 
            with open(filedir) as configfile:
                data = jsonmodule.load(configfile)
    
            data.update({self.name:savedata})
            with open(os.path.join(DATA_PATH, 'module.json') ,'w') as configfile:
                jsonmodule.dump(data, configfile, indent=4, sort_keys=True, 
                                cls=MyEncoder)
    
            print('Module {} updated in module.json'.format(self.name))
        # check that self.modulefile is not none
        if self.modulefile is None:
            self.modulefile = os.path.join('objects',
                                                   self.name + '.rad')
        
        if rewriteModulefile & os.path.isfile(self.modulefile):
            print(f"Pre-existing .rad file {self.modulefile} "
                  "will be overwritten\n")
            os.remove(self.modulefile)
            
        if not os.path.isfile(self.modulefile):
            # py2 and 3 compatible: binary write, encode text first
            with open(self.modulefile, 'wb') as f:
                f.write(self.text.encode('ascii'))
            
    def showModule(self):
        """ 
        Method to call objview and render the module object 
        (visualize it).
        
        Parameters: None

        """
       
        cmd = 'objview %s %s' % (os.path.join('materials', 'ground.rad'),
                                         self.modulefile)
        _,err = _popen(cmd,None)
        if err is not None:
            print('Error: {}'.format(err))
            print('possible solution: install radwinexe binary package from '
                  'http://www.jaloxa.eu/resources/radiance/radwinexe.shtml'
                  ' into your RADIANCE binaries path')
            return 

    def addTorquetube(self, diameter=0.1, tubetype='Round', material='Metal_Grey', 
                      axisofrotation=True, visible=True,  recompile=True):
        """
        For adding torque tubes to the module simulation. 
        
        Parameters
        ----------
        diameter : float   Tube diameter in meters. For square, diameter means 
                           the length of one of the square-tube side.  For Hex, 
                           diameter is the distance between two vertices 
                           (diameter of the circumscribing circle). Default 0.1
        tubetype : str     Options: 'Square', 'Round' (default), 'Hex' or 'Oct'
                           Tube cross section
        material : str     Options: 'Metal_Grey' or 'black'. Material for the 
                           torque tube.
        axisofrotation     (bool) :  Default True. IF true, creates geometry
                           so center of rotation is at the center of the 
                           torquetube, with an offsetfromaxis equal to half the
                           torquetube diameter + the zgap. If there is no 
                           torquetube (visible=False), offsetformaxis will 
                           equal the zgap.
        visible            (bool) :  Default True. If false, geometry is set
                           as if the torque tube were present (e.g. zgap, 
                           axisofrotation) but no geometry for the tube is made
        recompile : Bool          Rewrite .rad file and module.json file (default True)

        """
        self.torquetube = Tube(diameter=diameter, tubetype=tubetype,
                           material=material, axisofrotation=axisofrotation,
                           visible=visible)
        if recompile:
            self.compileText()


    def addOmega(self, omega_material='Metal_Grey', omega_thickness=0.004,
                 inverted=False, x_omega1=None, x_omega3=None, y_omega=None,
                 mod_overlap=None, recompile=True):
        """
        Add the racking structure element `omega`, which connects 
        the frame to the torque tube. 


        Parameters
        ----------

        omega_material : str      The material the omega structure is made of.
                                  Default: 'Metal_Grey'
        x_omega1  : float         The length of the module-adjacent arm of the 
                                  omega parallel to the x-axis of the module
        mod_overlap : float       The length of the overlap between omega and 
                                  module surface on the x-direction
        y_omega  : float          Length of omega (Y-direction)
        omega_thickness : float   Omega thickness. Default 0.004
        x_omega3  : float         X-direction length of the torquetube adjacent 
                                  arm of omega
        inverted : Bool           Modifies the way the Omega is set on the Torquetbue
                                  Looks like False: u  vs True: n  (default False)
                                  NOTE: The part that bridges the x-gap for a False
                                  regular orientation omega (inverted = False),
                                  is the x_omega3;
                                  and for inverted omegas (inverted=True) it is
                                  x_omega1.
        recompile : Bool          Rewrite .rad file and module.json file (default True)

        """
        self.omega = Omega(self, omega_material=omega_material,
                           omega_thickness=omega_thickness,
                           inverted=inverted, x_omega1=x_omega1,
                           x_omega3=x_omega3, y_omega=y_omega, 
                           mod_overlap=mod_overlap)
        if recompile:
            self.compileText()

    def addFrame(self, frame_material='Metal_Grey', frame_thickness=0.05, 
                 frame_z=0.3, nSides_frame=4, frame_width=0.05, recompile=True):
        """
        Add a metal frame geometry around the module.
        
        Parameters
        ------------

        frame_material : str    The material the frame structure is made of
        frame_thickness : float The profile thickness of the frame 
        frame_z : float         The Z-direction length of the frame that extends 
                                below the module plane
        frame_width : float     The length of the bottom frame that is bolted 
                                with the omega
        nSides_frame : int      The number of sides of the module that are framed.
                                4 (default) or 2

        """
        
        self.frame = Frame(frame_material=frame_material,
                   frame_thickness=frame_thickness,
                   frame_z=frame_z, nSides_frame=nSides_frame,
                   frame_width=frame_width)
        if recompile:
            self.compileText()
            
    def addCellModule(self, numcellsx, numcellsy ,xcell, ycell,
                      xcellgap=0.02, ycellgap=0.02, centerJB=None, recompile=True):
        """
        Create a cell-level module, with individually defined cells and gaps
        
        Parameters
        ------------
        numcellsx : int    Number of cells in the X-direction within the module
        numcellsy : int    Number of cells in the Y-direction within the module
        xcell : float      Width of each cell (X-direction) in the module
        ycell : float      Length of each cell (Y-direction) in the module
        xcellgap : float   Spacing between cells in the X-direction. 0.02 default
        ycellgap : float   Spacing between cells in the Y-direction. 0.02 default
        centerJB : float   (optional) Distance betwen both sides of cell arrays 
                           in a center-JB half-cell module. If 0 or not provided,
                           module will not have the center JB spacing. 
                           Only implemented for 'portrait' mode at the moment.
                           (numcellsy > numcellsx). 
        

        """
        import warnings
        if centerJB:
            warnings.warn(
                'centerJB functionality is currently experimental and subject '
                'to change in future releases. ' )
            
        
        self.cellModule = CellModule(numcellsx=numcellsx, numcellsy=numcellsy,
                                     xcell=xcell, ycell=ycell, xcellgap=xcellgap,
                                     ycellgap=ycellgap, centerJB=centerJB)
                                     
        if recompile:
            self.compileText()
    

    
    def _makeModuleFromDict(self,  x=None, y=None, z=None, xgap=None, ygap=None, 
                    zgap=None, numpanels=None, modulefile=None,
                    modulematerial=None, **kwargs):

        """
        go through and generate the text required to make a module
        """
        import warnings
        #aliases for equations below
        Ny = numpanels
        _cc = 0  # cc is an offset given to the module when cells are used
                  # so that the sensors don't fall in air when numcells is even.
                  # For non cell-level modules default is 0.
        # Update values for rotating system around torque tube.  
        diam=0
        if hasattr(self, 'torquetube'):
            diam = self.torquetube.diameter
            if self.torquetube.axisofrotation is True:
                self.offsetfromaxis = np.round(zgap + diam/2.0,8)
            if hasattr(self, 'frame'):
                self.offsetfromaxis = self.offsetfromaxis + self.frame.frame_z
        # TODO: make sure the above is consistent with old version below
        """
        if torquetube:
            diam = torquetube['diameter']
            torquetube_bool = torquetube['bool']
        else:
            diam=0
            torquetube_bool = False
        if self.axisofrotationTorqueTube == True:
            if torquetube_bool == True:
                self.offsetfromaxis = np.round(zgap + diam/2.0,8)
            else:
                self.offsetfromaxis = zgap
            if hasattr(self, 'frame'):
                self.offsetfromaxis = self.offsetfromaxis + self.frame.frame_z
        """
        # Adding the option to replace the module thickess
        if self.glass:
            zglass = 0.01
            print("\nWarning: module glass increases analysis variability. "  
                          "Recommend setting `accuracy='high'` in AnalysisObj.analysis().\n")
        else:
            zglass = 0.0
            
        if z is None:
            if self.glass:
                z = 0.001
            else:
                z = 0.020
                
        self.z = z
        self.zglass = zglass

            
        if modulematerial is None:
            modulematerial = 'black'
            self.modulematerial = 'black'
            
        if hasattr(self, 'cellModule'):
            (text, x, y, _cc) = self.cellModule._makeCellLevelModule(self, z, Ny, ygap, 
                                   modulematerial) 
        else:
            try:
                text = '! genbox {} {} {} {} {} '.format(modulematerial, 
                                                          self.name, x, y, z)
                text +='| xform -t {} {} {} '.format(-x/2.0,
                                        (-y*Ny/2.0)-(ygap*(Ny-1)/2.0),
                                        self.offsetfromaxis)
                text += '-a {} -t 0 {} 0'.format(Ny, y+ygap)
                packagingfactor = 100.0

            except Exception as err: # probably because no x or y passed
                raise Exception('makeModule variable {}'.format(err.args[0])+
                                ' and cellModule is None.  '+
                                'One or the other must be specified.')
 
            
        self.scenex = x + xgap
        self.sceney = np.round(y*numpanels + ygap*(numpanels-1), 8)
        self.scenez = np.round(zgap + diam / 2.0, 8)
        

        if hasattr(self, 'frame'):
            _zinc, frametext = self.frame._makeFrames( 
                                    x=x,y=y, ygap=ygap,numpanels=Ny, 
                                    offsetfromaxis=self.offsetfromaxis- 0.5*zglass)
        else:
            frametext = ''
            _zinc = 0  # z increment from frame thickness 
        _zinc = _zinc + 0.5 * zglass
            
        if hasattr(self, 'omega'):
            # This also defines scenex for length of the torquetube.
            omega2omega_x, omegatext = self.omega._makeOmega(x=x,y=y, xgap=xgap,  
                                            zgap=zgap, z_inc=_zinc, 
                                            offsetfromaxis=self.offsetfromaxis)
            if omega2omega_x > self.scenex:
                self.scenex =  omega2omega_x
            
            # TODO: is the above line better than below?
            #       I think this causes it's own set of problems, need to check.
            """
            if self.scenex <x:
                scenex = x+xgap #overwriting scenex to maintain torquetube continuity
        
                print ('Warning: Omega values have been provided, but' +
                       'the distance between modules with the omega'+
                       'does not match the x-gap provided.'+
                       'Setting x-gap to be the space between modules'+
                       'from the omega.')
            else:
                print ('Warning: Using omega-to-omega distance to define'+
                       'gap between modules'
                       +'xgap value not being used')
            """
        else:
            omegatext = ''
        
        # Defining scenex if it was not defined by the Omegas, 
        # after the module has been created in case it is a 
        # cell-level Module, in which the "x" gets calculated internally.
        # Also sanity check in case omega-to-omega distance is smaller
        # than module.

        #if torquetube_bool is True:
        if hasattr(self,'torquetube'):
            if self.torquetube.visible:
                text += self.torquetube._makeTorqueTube(cc=_cc, zgap=zgap,   
                                         z_inc=_zinc, scenex=self.scenex)

        # TODO:  should there be anything updated here like scenez?
        #        YES.
        if self.glass: 
                edge = 0.01                     
                text = text+'\r\n! genbox stock_glass {} {} {} {} '.format(self.name+'_Glass',x+edge, y+edge, zglass)
                text +='| xform -t {} {} {} '.format(-x/2.0-0.5*edge + _cc,
                                        (-y*Ny/2.0)-(ygap*(Ny-1)/2.0)-0.5*edge,
                                        self.offsetfromaxis - 0.5*zglass)
                text += '-a {} -t 0 {} 0'.format(Ny, y+ygap)
            

        text += frametext
        if hasattr(self, 'omega'):
            text += self.omega.text    
        text += self.customtext  # For adding any other racking details at the module level that the user might want.

        self.text = text
        return text
    #End of makeModuleFromDict()
  
# end of ModuleObj



class Omega(SuperClass):

    def __init__(self, module, omega_material='Metal_Grey', omega_thickness=0.004,
                 inverted=False, x_omega1=None, x_omega3=None, y_omega=None,
                 mod_overlap=None):
        """
        ====================    ===============================================
        Keys : type             Description
        ================        =============================================== 
        module : ModuleObj      Parent object with details related to geometry
        omega_material : str    The material the omega structure is made of
        omega_thickness : float Omega thickness
        inverted : Bool         Modifies the way the Omega is set on the Torquetbue
                                Looks like False: u  vs True: n  (default False)
        x_omega1  : float       The length of the module-adjacent arm of the 
                                omega parallel to the x-axis of the module
        y_omega  : float         Length of omega (Y-direction)
        x_omega3  : float       X-direction length of the torquetube adjacent 
                                arm of omega
        mod_overlap : float     The length of the overlap between omega and 
                                module surface on the x-direction

        =====================   ===============================================

        """
        self.keys = ['omega_material', 'x_omega1', 'mod_overlap', 'y_omega', 
            'omega_thickness','x_omega3','inverted']

        if x_omega1 is None:
            if inverted:
                x_omega1 = module.xgap*0.5
            else:
                x_omega1 = module.xgap*0.5*0.6
            _missingKeyWarning('Omega', 'x_omega1', x_omega1)
                
        if x_omega3 is None:
            if inverted:
                x_omega3 = module.xgap*0.5*0.3
            else:
                x_omega3 = module.xgap*0.5
            _missingKeyWarning('Omega', 'x_omega3', x_omega3)
            
        if y_omega is None:
            y_omega = module.y/2
            _missingKeyWarning('Omega', 'y_omega', y_omega)
        
        if mod_overlap is None:
           mod_overlap = x_omega1*0.6
           _missingKeyWarning('Omega', 'mod_overlap', mod_overlap)

        # set data object attributes from datakey list. 
        for key in self.keys:
            setattr(self, key, eval(key))        

        
        
    def _makeOmega(self, x, y, xgap, zgap, offsetfromaxis, z_inc = 0, **kwargs):
        """
        Helper function for creating a module that includes the racking 
        structure element `omega`.  

        TODO: remove some or all of this documentation since this is an internal function    
        
        Parameters
        ------------
        x : numeric
            Width of module along the axis of the torque tube or racking structure. (meters).
        y : numeric
            Length of module (meters)
        xgap : float
            Panel space in X direction. Separation between modules in a row.
        zgap : float
            Distance behind the modules in the z-direction to the edge of the tube (m)
        offsetfromaxis : float
            Internally defined variable in makeModule that specifies how much
            the module is offset from the Axis of Rotation due to zgap and or 
            frame thickness.
        z_inc : dict
            Internally defined variable in makeModule that specifies how much
            the module is offseted by the Frame.
        
        """
        
        # set local variables
        omega_material = self.omega_material
        x_omega1 = self.x_omega1
        mod_overlap = self.mod_overlap
        y_omega = self.y_omega
        omega_thickness = self.omega_thickness
        x_omega3 = self.x_omega3

        
        z_omega2 = zgap
        x_omega2 = omega_thickness 
        z_omega1 = omega_thickness
        z_omega3 = omega_thickness
        
        #naming the omega pieces
        name1 = 'mod_adj'
        name2 = 'verti'
        name3 = 'tt_adj'
        
        
        # defining the module adjacent member of omega
        x_translate1 = -x/2 - x_omega1 + mod_overlap
        y_translate = -y_omega/2 #common for all the pieces
        z_translate1 = offsetfromaxis-z_omega1
        
        #defining the vertical (zgap) member of the omega
        x_translate2 = x_translate1
        z_translate2 = offsetfromaxis-z_omega2
            
        #defining the torquetube adjacent member of omega
        x_translate3 = x_translate1-x_omega3
        z_translate3 = z_translate2
        
        if z_inc != 0: 
            z_translate1 += -z_inc
            z_translate2 += -z_inc
            z_translate3 += -z_inc
        
        # for this code, only the translations need to be shifted for the inverted omega
        
        if self.inverted == True:
            # shifting the non-inv omega shape of west as inv omega shape of east
            x_translate1_inv_east = x/2-mod_overlap
            x_shift_east = x_translate1_inv_east - x_translate1

            # shifting the non-inv omega shape of west as inv omega shape of east
            x_translate1_inv_west = -x_translate1_inv_east - x_omega1
            x_shift_west = -x_translate1_inv_west + (-x_translate1-x_omega1)
            
            #customizing the East side of the module for omega_inverted

            omegatext = '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name1, x_omega1, y_omega, z_omega1, x_translate1_inv_east, y_translate, z_translate1) 
            omegatext += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name2, x_omega2, y_omega, z_omega2, x_translate2 + x_shift_east, y_translate, z_translate2)
            omegatext += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name3, x_omega3, y_omega, z_omega3, x_translate3 + x_shift_east, y_translate, z_translate3)

            #customizing the West side of the module for omega_inverted

            omegatext += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name1, x_omega1, y_omega, z_omega1, x_translate1_inv_west, y_translate, z_translate1) 
            omegatext += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name2, x_omega2, y_omega, z_omega2, -x_translate2-x_omega2 -x_shift_west, y_translate, z_translate2)
            omegatext += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name3, x_omega3, y_omega, z_omega3, -x_translate3-x_omega3 - x_shift_west, y_translate, z_translate3)
            
            omega2omega_x = -x_translate1_inv_east*2
        
        else:
            
            #customizing the West side of the module for omega
            
            omegatext = '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name1, x_omega1, y_omega, z_omega1, x_translate1, y_translate, z_translate1) 
            omegatext += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name2, x_omega2, y_omega, z_omega2, x_translate2, y_translate, z_translate2)
            omegatext += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name3, x_omega3, y_omega, z_omega3, x_translate3, y_translate, z_translate3)
                
            #customizing the East side of the module for omega
                
            omegatext += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name1, x_omega1, y_omega, z_omega1, -x_translate1-x_omega1, y_translate, z_translate1) 
            omegatext += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name2, x_omega2, y_omega, z_omega2, -x_translate2-x_omega2, y_translate, z_translate2)
            omegatext += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(omega_material, name3, x_omega3, y_omega, z_omega3, -x_translate3-x_omega3, y_translate, z_translate3)
        
            omega2omega_x = -x_translate3*2
        self.text = omegatext
        self.omega2omega_x = omega2omega_x
        return omega2omega_x,omegatext
    
class Frame(SuperClass):

    def __init__(self, frame_material='Metal_Grey', frame_thickness=0.05, 
                 frame_z=None, nSides_frame=4, frame_width=0.05):
        """
        Parameters
        ------------

        frame_material : str    The material the frame structure is made of
        frame_thickness : float The profile thickness of the frame 
        frame_z : float         The Z-direction length of the frame that extends 
                                below the module plane
        frame_width : float     The length of the bottom frame that is bolted 
                                with the omega
        nSides_frame : int      The number of sides of the module that are framed.
                                4 (default) or 2


        """
        self.keys = ['frame_material', 'frame_thickness', 'frame_z', 'frame_width',
            'nSides_frame']
        
        if frame_z is None:
            frame_z = 0.03
            _missingKeyWarning('Frame', 'frame_z', frame_z)
        
        # set data object attributes from datakey list. 
        for key in self.keys:
            setattr(self, key, eval(key))  
        
    def _makeFrames(self,  x, y, ygap, numpanels, offsetfromaxis):
        """
        Helper function for creating a module that includes the frames attached to the module, 

            
        Parameters
        ------------
        frameParams : dict
            Dictionary with input parameters for creating a frame as part of the module.
            See details below for keys needed.
        x : numeric
            Width of module along the axis of the torque tube or racking structure. (meters).
        y : numeric
            Length of module (meters)
        ygap : float
            Gap between modules arrayed in the Y-direction if any.
        numpanels : int
            Number of modules arrayed in the Y-direction. e.g.
            1-up or 2-up, etc. (supports any number for carport/Mesa simulations)
        offsetfromaxis : float
            Internally defined variable in makeModule that specifies how much
            the module is offset from the Axis of Rotation due to zgap and or 
            frame thickness.


        """
        
        # 
        if self.nSides_frame == 2 and x>y:
            print("Development Warning: Frames has only 2 sides and module is"+
                  "in ladscape. This functionality is not working properly yet"+
                  "for this release. We are overwriting nSide_frame = 4 to continue."+
                  "If this functionality is pivotal to you we can prioritize adding it but"+
                  "please comunicate with the development team. Thank you.")
            self.nSides_frame = 4
        
        #Defining internal names
        frame_material = self.frame_material 
        f_thickness = self.frame_thickness 
        f_height = self.frame_z
        n_frame = self.nSides_frame  
        fl_x = self.frame_width

        y_trans_shift = 0 #pertinent to the case of x>y with 2-sided frame
                

        # Recalculating width ignoring the thickness of the aluminum
        # for internal positioining and sizing of hte pieces
        fl_x = fl_x-f_thickness
        
        if x>y and n_frame==2:
            x_temp,y_temp = y,x
            rotframe = 90
            frame_y = x
            y_trans_shift = x/2-y/2
        else:
            x_temp,y_temp = x,y
            frame_y = y
            rotframe = 0
    
        Ny = numpanels
        y_half = (y*Ny/2)+(ygap*(Ny-1)/2)
    
        # taking care of lengths and translation points
        # The pieces are same and symmetrical for west and east
    
        # naming the frame pieces
        nameframe1 = 'frameside'
        nameframe2 = 'frameleg'
        
        #frame sides
        few_x = f_thickness
        few_y = frame_y
        few_z = f_height
    
        fw_xt = -x_temp/2 # in case of x_temp = y this doesn't reach panel edge
        fe_xt = x_temp/2-f_thickness 
        few_yt = -y_half-y_trans_shift
        few_zt = offsetfromaxis-f_height
    
        #frame legs for east-west 
    
        flw_xt = -x_temp/2 + f_thickness
        fle_xt = x_temp/2 - f_thickness-fl_x
        flew_yt = -y_half-y_trans_shift
        flew_zt = offsetfromaxis-f_height
    
    
        #pieces for the shorter side (north-south in this case)
    
        #filler
    
        fns_x = x_temp-2*f_thickness
        fns_y = f_thickness
        fns_z = f_height-f_thickness
    
        fns_xt = -x_temp/2+f_thickness
        fn_yt = -y_half+y-f_thickness
        fs_yt = -y_half
        fns_zt = offsetfromaxis-f_height+f_thickness
    
        # the filler legs
    
        filleg_x = x_temp-2*f_thickness-2*fl_x
        filleg_y = f_thickness + fl_x
        filleg_z = f_thickness
    
        filleg_xt = -x_temp/2+f_thickness+fl_x
        fillegn_yt = -y_half+y-f_thickness-fl_x
        fillegs_yt = -y_half
        filleg_zt = offsetfromaxis-f_height
    
    
        # making frames: west side
        
        
        frame_text = '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(frame_material, nameframe1, few_x, few_y, few_z, fw_xt, few_yt, few_zt) 
        frame_text += ' -a {} -t 0 {} 0 | xform -rz {}'.format(Ny, y_temp+ygap, rotframe)
    
        frame_text += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(frame_material, nameframe2, fl_x, frame_y, f_thickness, flw_xt, flew_yt, flew_zt)
        frame_text += ' -a {} -t 0 {} 0 | xform -rz {}'.format(Ny, y_temp+ygap, rotframe)
                
        # making frames: east side
    
        frame_text += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(frame_material, nameframe1, few_x, few_y, few_z, fe_xt, few_yt, few_zt) 
        frame_text += ' -a {} -t 0 {} 0 | xform -rz {}'.format(Ny, y_temp+ygap, rotframe)
    
        frame_text += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(frame_material, nameframe2, fl_x, frame_y, f_thickness, fle_xt, flew_yt, flew_zt)
        frame_text += ' -a {} -t 0 {} 0 | xform -rz {}'.format(Ny, y_temp+ygap, rotframe)

    
        if n_frame == 4:
            #making frames: north side
    
            frame_text += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(frame_material, nameframe1, fns_x, fns_y, fns_z, fns_xt, fn_yt, fns_zt) 
            frame_text += ' -a {} -t 0 {} 0'.format(Ny, y+ygap)
    
    
            frame_text += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(frame_material, nameframe2, filleg_x, filleg_y, filleg_z, filleg_xt, fillegn_yt, filleg_zt)
            frame_text += ' -a {} -t 0 {} 0'.format(Ny, y+ygap)
    
            #making frames: south side
    
            frame_text += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(frame_material, nameframe1, fns_x, fns_y, fns_z, fns_xt, fs_yt, fns_zt) 
            frame_text += ' -a {} -t 0 {} 0'.format(Ny, y+ygap)
    
            frame_text += '\r\n! genbox {} {} {} {} {} | xform -t {} {} {}'.format(frame_material, nameframe2, filleg_x, filleg_y, filleg_z, filleg_xt, fillegs_yt, filleg_zt)
            frame_text += ' -a {} -t 0 {} 0'.format(Ny, y+ygap)

        z_inc = f_height

        return z_inc, frame_text

class Tube(SuperClass):

    def __init__(self, diameter=0.1, tubetype='Round', material='Metal_Grey', 
                      axisofrotation=True, visible=True):
        """
        ================   ====================================================
        Keys : type        Description
        ================   ====================================================  
        diameter : float   Tube diameter in meters. For square, diameter means 
                           the length of one of the square-tube side.  For Hex, 
                           diameter is the distance between two vertices 
                           (diameter of the circumscribing circle). Default 0.1
        tubetype : str     Options: 'Square', 'Round' (default), 'Hex' or 'Oct'
                           Tube cross section
        material : str     Options: 'Metal_Grey' or 'black'. Material for the 
                           torque tube.
        axisofrotation     (bool) :  Default True. IF true, creates geometry
                           so center of rotation is at the center of the 
                           torquetube, with an offsetfromaxis equal to half the
                           torquetube diameter + the zgap. If there is no 
                           torquetube (visible=False), offsetformaxis will 
                           equal the zgap.
        visible            (bool) :  Default True. If false, geometry is set
                           as if the torque tube were present (e.g. zgap, 
                           axisofrotation) but no geometry for the tube is made
        ================   ==================================================== 
        """
        
        self.keys = ['diameter', 'tubetype', 'material', 'visible']   # what about axisofrotation?
        
        self.axisofrotation = axisofrotation
        # set data object attributes from datakey list. 
        for key in self.keys:
            setattr(self, key, eval(key))    
            
    def _makeTorqueTube(self, cc,  z_inc, zgap, scenex):
        """  
        Return text string for generating the torque tube geometry
        
        Parameters
        
        cc = module._cc #horizontal offset to center of a cell
        """
        import math
        
        
        text = ''
        tto = 0  # Torquetube Offset. Default = 0 if axisofrotationTT == True
        diam = self.diameter  #alias
        material = self.material #alias
        
        if self.tubetype.lower() == 'square':
            if self.axisofrotation == False:
                tto = -z_inc-zgap-diam/2.0
            text += '\r\n! genbox {} tube1 {} {} {} '.format(material,
                                  scenex, diam, diam)
            text += '| xform -t {} {} {}'.format(-(scenex)/2.0+cc,
                                -diam/2.0, -diam/2.0+tto)

        elif self.tubetype.lower() == 'round':
            if self.axisofrotation == False:
                tto = -z_inc-zgap-diam/2.0
            text += '\r\n! genrev {} tube1 t*{} {} '.format(material, scenex, diam/2.0)
            text += '32 | xform -ry 90 -t {} {} {}'.format(-(scenex)/2.0+cc, 0, tto)

        elif self.tubetype.lower() == 'hex':
            radius = 0.5*diam

            if self.axisofrotation == False:
                tto = -z_inc-radius*math.sqrt(3.0)/2.0-zgap

            text += '\r\n! genbox {} hextube1a {} {} {} | xform -t {} {} {}'.format(
                    material, scenex, radius, radius*math.sqrt(3),
                    -(scenex)/2.0+cc, -radius/2.0, -radius*math.sqrt(3.0)/2.0+tto) #ztran -radius*math.sqrt(3.0)-tto


            # Create, translate to center, rotate, translate back to prev. position and translate to overal module position.
            text = text+'\r\n! genbox {} hextube1b {} {} {} | xform -t {} {} {} -rx 60 -t 0 0 {}'.format(
                    material, scenex, radius, radius*math.sqrt(3), -(scenex)/2.0+cc, -radius/2.0, -radius*math.sqrt(3.0)/2.0, tto) #ztran (radius*math.sqrt(3.0)/2.0)-radius*math.sqrt(3.0)-tto)
            
            text = text+'\r\n! genbox {} hextube1c {} {} {} | xform -t {} {} {} -rx -60 -t 0 0 {}'.format(
                    material, scenex, radius, radius*math.sqrt(3), -(scenex)/2.0+cc, -radius/2.0, -radius*math.sqrt(3.0)/2.0, tto) #ztran (radius*math.sqrt(3.0)/2.0)-radius*math.sqrt(3.0)-tto)

        elif self.tubetype.lower()=='oct':
            radius = 0.5*diam
            s = diam / (1+math.sqrt(2.0))   # 

            if self.axisofrotation == False:
                tto = -z_inc-radius-zgap

            text = text+'\r\n! genbox {} octtube1a {} {} {} | xform -t {} {} {}'.format(
                    material, scenex, s, diam, -(scenex)/2.0, -s/2.0, -radius+tto)

            # Create, translate to center, rotate, translate back to prev. position and translate to overal module position.
            text = text+'\r\n! genbox {} octtube1b {} {} {} | xform -t {} {} {} -rx 45 -t 0 0 {}'.format(
                    material, scenex, s, diam, -(scenex)/2.0+cc, -s/2.0, -radius, tto)

            text = text+'\r\n! genbox {} octtube1c {} {} {} | xform -t {} {} {} -rx 90 -t 0 0 {}'.format(
                    material, scenex, s, diam, -(scenex)/2.0+cc, -s/2.0, -radius, tto)

            text = text+'\r\n! genbox {} octtube1d {} {} {} | xform -t {} {} {} -rx 135 -t 0 0 {} '.format(
                    material, scenex, s, diam, -(scenex)/2.0+cc, -s/2.0, -radius, tto)


        else:
            raise Exception("Incorrect torque tube type.  "+
                            "Available options: 'square' 'oct' 'hex' or 'round'."+
                            "  Value entered: {}".format(self.tubetype))    
        self.text = text
        return text
    
class CellModule(SuperClass):

    def __init__(self, numcellsx, numcellsy,
                 xcell, ycell, xcellgap=0.02, ycellgap=0.02, centerJB=None):
        """
        For creating a cell-level module, the following input parameters should 
        be in ``cellModule``:
        
        ================   ====================================================
        Keys : type        Description
        ================   ====================================================  
        numcellsx : int    Number of cells in the X-direction within the module
        numcellsy : int    Number of cells in the Y-direction within the module
        xcell : float      Width of each cell (X-direction) in the module
        ycell : float      Length of each cell (Y-direction) in the module
        xcellgap : float   Spacing between cells in the X-direction. 0.02 default
        ycellgap : float   Spacing between cells in the Y-direction. 0.02 default
        centerJB : float   (optional) Distance betwen both sides of cell arrays 
                           in a center-JB half-cell module. If 0 or not provided,
                           module will not have the center JB spacing. 
                           Only implemented for 'portrait' mode at the moment.
                           (numcellsy > numcellsx). 
        cc : float         center cell offset from x so scan is not at a gap 
                           between cells
        ================   ==================================================== 

        """
        self.keys = ['numcellsx', 'numcellsy', 'xcell', 'ycell', 'xcellgap',
            'ycellgap','centerJB'] 
        
        # set data object attributes from datakey list. 
        for key in self.keys:
            setattr(self, key, eval(key))    
        
    
    def _makeCellLevelModule(self, module, z, Ny, ygap, 
                         modulematerial):
        """  Calculate the .radfile generation text for a cell-level module.
        """
        offsetfromaxis = module.offsetfromaxis
        c = self.getDataDict()

        # For half cell modules with the JB on the center:
        if c['centerJB'] is not None:
            centerJB = c['centerJB']
            y = c['numcellsy']*c['ycell'] + (c['numcellsy']-2)*c['ycellgap'] + centerJB            
        else:
            centerJB = 0
            y = c['numcellsy']*c['ycell'] + (c['numcellsy']-1)*c['ycellgap']

        x = c['numcellsx']*c['xcell'] + (c['numcellsx']-1)*c['xcellgap']

        #center cell -
        if c['numcellsx'] % 2 == 0:
            _cc = c['xcell']/2.0
            print("Module was shifted by {} in X to avoid sensors on air".format(_cc))
        else:
            _cc = 0

        text = '! genbox {} cellPVmodule {} {} {} | '.format(modulematerial,
                                               c['xcell'], c['ycell'], z)
        text +='xform -t {} {} {} '.format(-x/2.0 + _cc,
                         (-y*Ny / 2.0)-(ygap*(Ny-1) / 2.0)-centerJB/2.0,
                         offsetfromaxis)
        
        text += '-a {} -t {} 0 0 '.format(c['numcellsx'], c['xcell'] + c['xcellgap'])
        
        if centerJB != 0:
            trans0 = c['ycell'] + c['ycellgap']
            text += '-a {} -t 0 {} 0 '.format(c['numcellsy']/2, trans0)
            #TODO: Continue playing with the y translation of the array in the next two lines
                 # Until it matches. Close but not there.
            # This is 0 spacing
            #ytrans1 = y/2.0-c['ycell']/2.0-c['ycellgap']+centerJB/2.0   # Creating the 2nd array with the right Jbox distance
            ytrans1 = y/2.0-c['ycell']/2.0-c['ycellgap']+centerJB/2.0 + centerJB
            ytrans2= c['ycell'] - centerJB/2.0 + c['ycellgap']/2.0
            text += '-a {} -t 0 {} 0 '.format(2, ytrans1)  
            text += '| xform -t 0 {} 0 '.format(ytrans2)   

        else:
            text += '-a {} -t 0 {} 0 '.format(c['numcellsy'], c['ycell'] + c['ycellgap'])
            
        text += '-a {} -t 0 {} 0'.format(Ny, y+ygap)

        # OPACITY CALCULATION
        packagingfactor = np.round((c['xcell']*c['ycell']*c['numcellsx']*c['numcellsy'])/(x*y), 2)
        print("This is a Cell-Level detailed module with Packaging "+
              "Factor of {} %".format(packagingfactor)) 
        
        module.x = x
        module.y = y
        self.text = text
        
        return(text, x, y, _cc)    

# deal with Int32 JSON incompatibility
# https://www.programmerall.com/article/57461489186/
import json
class MyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(MyEncoder, self).default(obj)


from setuptools import setup,find_packages,Extension
from setuptools import find_packages
from pathlib import Path

# # Load the README file.
here = Path(__file__).parent

file = open("Readme.md", encoding="utf8")
long_description = file.read()

    
setup(
    

    # Define the library name, this is what is used along with `pip install`.
    name='BifacialSimu',

    # Define the author of the repository.
    author='Eva-Maria Grommes',

    # Define the Author's email, so people know who to reach out to.
    author_email='Eva-Maria.Grommes@th-koeln.de',

    # Define the version of this library.
    # Read this as
    #   - MAJOR VERSION 1
    #   - MINOR VERSION 0
    #   - MAINTENANCE VERSION 0
    version='1.0.2',

    # Here is a small description of the library. This appears
    # when someone searches for the library on https://pypi.org/search.
    description='Holistic Simulation of large-scale Bifacial Photovoltaic Systems',

    # I have a long description but that will just be my README
    # file, note the variable up above where I read the file.
    long_description=long_description,

    # This will specify that the long description is MARKDOWN.
    long_description_content_type='text/markdown',

    # Here is the URL where you can find the code, in this case on GitHub.
    url='https://github.com/cire-thk/BifacialSimu',

    # These are the dependencies the library needs in order to run.
    install_requires=[
        'pandas==1.3.4',
        'ipython==8.4.0',
        'matplotlib==3.4.3',
        'numpy==1.20.3',
        'Pillow==9.1.1',
        'pvlib==0.9.1',
        'python_dateutil==2.8.2',
        'pytz==2021.3',
        'requests==2.26.0',
        'seaborn==0.11.2',
        'Shapely==1.8.2',
        'tqdm==4.62.3',
        'bifacial_radiance==0.4.1',#content changed!
        # 'pyqt5==5.12.3'
        # 'pyqtwebengine==5.12.1'
        'pvfactors==1.5.2'
    ],

    # Here are the keywords of my library.
    keywords='Bifacialsimu, bifacial simulation,  energy yield of bifacial PV systems, albedo',

    # here are the packages I want "build."
    packages=['BifacialSimu_src'],
    
    # I also have some package data, like photos and JSON files, so
    # I want to include those as well.
    include_package_data=True,
      
    classifiers=[
           # Here I'll specify the version of Python it uses.
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.9',

          ]
# ==================== ADDTIONAL PACKAGE INFO =================================
# 
#     # # here we specify any package data.
#     # package_data={
# 
#     #     # And include any files found subdirectory of the "td" package.
#     #     "td": ["app/*", "templates/*"],
# 
#     # },
# 
#     # I also have some package data, like photos and JSON files, so
#     # I want to include those as well.
#     include_package_data=True,
# 
#     # Here I can specify the python version necessary to run this library.
#     python_requires='>=3.7',
# 
#     # Additional classifiers that give some characteristics about the package.
#     # For a complete list go to https://pypi.org/classifiers/.
#     classifiers=[
# 
#         # I can say what phase of development my library is in.
#         'Development Status :: 3 - Alpha',
# 
#         # Here I'll add the audience this library is intended for.
#         'Intended Audience :: Developers',
#         'Intended Audience :: Science/Research',
#         'Intended Audience :: Financial and Insurance Industry',
# 
#         # Here I'll define the license that guides my library.
#         'License :: OSI Approved :: MIT License',
# 
#         # Here I'll note that package was written in English.
#         'Natural Language :: English',
# 
#         # Here I'll note that any operating system can use it.
#         'Operating System :: OS Independent',
# 
#         # Here I'll specify the version of Python it uses.
#         'Programming Language :: Python',
#         'Programming Language :: Python :: 3',
#         'Programming Language :: Python :: 3.8',
# 
#         # Here are the topics that my library covers.
#         'Topic :: Database',
#         'Topic :: Education',
#         'Topic :: Office/Business'
# 
#     ]
# =============================================================================
)


from .BifacialSimu_calculationHandler import *
from .BifacialSimu_dataHandler import *
from .BifacialSimu_radiationHandler import *
from .BifacialSimu_spectralAlbedoHandler import *
from .BifacialSimu_spectralAlbedoHandler_1_row import *


# -*- coding: utf-8 -*-
"""
Created on Mon Jun  7 11:39:16 2021
@author:        
    Eva-Maria Grommes

Additional co-authors can be found here:
https://github.com/cire-thk/bifacialSimu    

name:
    BifacialSimu - dataHandler

overview:
    Handles the import and export of data 
    used by the bifacial simulation of PV-Modules

"""

# Import/transformation of needed data
# Import/transformation of Excel sheets
    # function to merge radiation data
    # function to export radiation data
    # function to export graphics and plots
    # function to create reports


import datetime
import pandas as pd
import os #to import directories
import sys
import dateutil
import numpy as np






# Path handling
rootPath = os.path.dirname(os.path.dirname(os.path.realpath(".")))

from . import BifacialSimu_radiationHandler #"from ." is essential to tell python that the module is found in this file's PATH

class DataHandler:
    def __init__(self):
        self.localDir = rootPath

    
    def setDirectories(self, outputFolder = "Outputs"):
        '''
        outputFolder: str
            Name of output folder (will be created)
        
        '''
  
        # Path to import irradiance data / Quality for plot export
        now = datetime.datetime.now()
        date_time = now.strftime("%Y %m %d_%H_%M") # get current date and time
        outputPath = os.path.join(self.localDir, outputFolder)
        resultsPath = os.path.join(outputPath, date_time + '_results/' ) 
        if not os.path.exists(resultsPath):
            os.makedirs(resultsPath)        # create path to output
        
        return resultsPath
    
    
    def getWeatherData(self, simulationDict, resultsPath):
        """
        Function to create a Radiance Obj with bifacial_radiance and read weather data.
        Can read EPW weather files from input location data or local weather files
        
        Parameters
        ----------
        simulationDict: simulation Dictionary, which can be found in BifacialSimuu_main.py
        resultsPath: output filepath
        """
        
        demo = BifacialSimu_radiationHandler.RayTrace.createDemo(simulationDict, resultsPath)
        if simulationDict['localFile'] == False:
            try:
                longitude = simulationDict['longitude']
                latitude = simulationDict['latitude']
            
                epwfile = demo.getEPW(latitude,longitude) # pull TMY data for any global lat/lon
            except ConnectionError: # no connection to automatically pull data
                pass
            metdata = demo.readEPW(epwfile) # read in the EPW weather data from above
        else:
            metdata = demo.readTMY(simulationDict['weatherFile'])
        
        return metdata, demo
    
    def passEPWtoDF(self, metdata, simulationDict, resultsPath):
        
        """
        Function to pass irradiance data and temperature from metdata Object created by bifacial_radiance to a pandas dataframe.
        The dataframe will be further used in the class ViewFactors.
        Additionally, a timeindex will be set.
        
        Parameters
        ----------
        simulationDict: simulation Dictionary, which can be found in BifacialSimu_main.py
        metdata: Object containing meteorological data and sun parameters        
        resultsPath: output filepath       
        """
        df = metdata.solpos

        df['ghi'] = metdata.ghi
        df['dhi'] = metdata.dhi
        df['dni'] = metdata.dni

        df['temperature'] = metdata.temp_air
        df['pressure'] = metdata.pressure
        df['albedo'] = metdata.albedo
        
        #define start and end Date for dataframe
        #dtStart = datetime.datetime(simulationDict['startHour'][0], simulationDict['startHour'][1], simulationDict['startHour'][2], simulationDict['startHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
        #dtEnd = datetime.datetime(simulationDict['endHour'][0], simulationDict['endHour'][1], simulationDict['endHour'][2], simulationDict['endHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
         
       
        df = df.reset_index()
        
        
        df['corrected_timestamp'] = pd.to_datetime(df['corrected_timestamp'])
        #add 30 minutes, since the calculation of the sunposition has changed the dateformat to -30 minutes 
        df['corrected_timestamp'] = df['corrected_timestamp'] + datetime.timedelta(minutes=30)
        
        #change the days at midnight, because through the shifting it is not right anymore
        if simulationDict['localFile'] == True:
            
            df['corrected_timestamp'] = df['corrected_timestamp'].astype(str)
            df['is_midnight']= df['corrected_timestamp'].str[11:13].apply(lambda x: 'YES' if (x == '00') else 'NO') 
            df['corrected_timestamp'] = pd.to_datetime(df['corrected_timestamp'])
            df['corrected_timestamp'] = np.where(df['is_midnight'] == "YES", df['corrected_timestamp'] + datetime.timedelta(days=-1), df['corrected_timestamp'])
            df.drop(columns=['is_midnight'])
            
        df = df.set_index('corrected_timestamp')
        
        #check if the year of the weatherfile is the same as which the user has entered
        if str(df.index.year[0]) != str(simulationDict['startHour'][0]):
              sys.exit("Please correct the input year to match the year of the weather file: " + str(df.index.year[0]))
            
        
        print('view_factor dataframe at data handler:')
        print(df)

        df.to_csv(resultsPath + "Dataframe_df.csv")
        return df


# -*- coding: utf-8 -*-
"""
Created on Tue Mar 26 20:16:47 2019

@author: sayala
"""

#from load import * 



def _sensorupsampletocellsbyInterpolation(df, cellsy):
    '''
        
    Function for when sensorsy in the results are less than cellsy desired.
    Interpolates the dataframe.
    
    #2DO: improve interpolation with pandas. right onw it's row by row.
    
    _sensorupsampletocellsbyInterpolation(df, cellsy)
    '''
    
    import pandas as pd
    import numpy as np
    
    sensorsy = len(df)
    
    #2DO: Update this section to match bifacialvf
    cellCenterPVM=[]                                    
    for i in range (0, cellsy):
        cellCenterPVM.append((i*sensorsy/cellsy+(i+1)*sensorsy/cellsy)/2)
    
    df2 = pd.DataFrame()
    for j in range (0, len(df.keys())):
        A = list(df[df.keys()[j]])
        B= np.interp(cellCenterPVM, list(range(0,sensorsy)), A)
        df2[df.keys()[j]]=B
    
    return df2
    
def _sensorsdownsampletocellsbyAverage(df, cellsy):
    '''
    df = dataframe with rows indexed by number (i.e. 0 to sensorsy) where sensorsy > cellsy
    cellsy = int. usually 8 or 12.

    example:
    F_centeraverages = _sensorsdownsampletocellsbyAverage(F, cellsy)
    '''
    import numpy as np
    import pandas as pd

    edges=len(df)-np.floor(len(df)/(cellsy))*(cellsy)
    edge1=int(np.floor(edges/2))
    edge2=int(edges-edge1)
    A = list(range(df.index[0]+edge1, df.index[-1]-edge2+2, int(np.floor(len(df)/(cellsy)))))
    B = range(0,len(A)-1,1)
    C = [df.iloc[A[x]:A[x+1]].mean(axis=0) for x in B]
    df_centeraverages=pd.DataFrame(C)
    
    return df_centeraverages

def _sensorsdownsampletocellbyCenter(df, cellsy):
    '''
    df = dataframe with rows indexed by number (i.e. 0 to sensorsy) where sensorsy > cellsy
    cellsy = int. usually 8 or 12.

    example:
    F_centervalues = _sensorsdownsampletocellbyCenter(F, cellsy)
    '''

    import numpy as np

    
    edges=len(df)-np.floor(len(df)/(cellsy))*(cellsy)
    edge1=int(np.floor(edges/2))
    edge2=int(edges-edge1)
    A = list(range(df.index[0]+edge1, df.index[-1]-edge2+2, int(np.floor(len(df)/(cellsy)))))
    A = [int(x+(A[1]-A[0])*0.5) for x in A]
    A = A[:-1]
    df_centervalues=df.loc[A]
    df_centervalues=df_centervalues.reset_index(drop=True)
    
    return df_centervalues
    
def _setupforPVMismatch(portraitorlandscape, sensorsy, numcells=72):
    r''' Sets values for calling PVMismatch, for ladscape or portrait modes and 
    
    Example:
    stdpl, cellsx, cellsy = _setupforPVMismatch(portraitorlandscape='portrait', sensorsy=100):
    '''

    import numpy as np

        # cell placement for 'portrait'.
    if numcells == 72:
        stdpl=np.array([[0,	23,	24,	47,	48,	71],
        [1,	22,	25,	46,	49,	70],
        [2,	21,	26,	45,	50,	69],
        [3,	20,	27,	44,	51,	68],
        [4,	19,	28,	43,	52,	67],
        [5,	18,	29,	42,	53,	66],
        [6,	17,	30,	41,	54,	65],
        [7,	16,	31,	40,	55,	64],
        [8,	15,	32,	39,	56,	63],
        [9,	14,	33,	38,	57,	62],
        [10,	13,	34,	37,	58,	61],
        [11,	12,	35,	36,	59,	60]])
    
    elif numcells == 96:
        stdpl=np.array([[0,	23,	24,	47,	48,	71,	72,	95],
            [1,	22,	25,	46,	49,	70,	73,	94],
            [2,	21,	26,	45,	50,	69,	74,	93],
            [3,	20,	27,	44,	51,	68,	75,	92],
            [4,	19,	28,	43,	52,	67,	76,	91],
            [5,	18,	29,	42,	53,	66,	77,	90],
            [6,	17,	30,	41,	54,	65,	78,	89],
            [7,	16,	31,	40,	55,	64,	79,	88],
            [8,	15,	32,	39,	56,	63,	80,	87],
            [9,	14,	33,	38,	57,	62,	81,	86],
            [10,	13,	34,	37,	58,	61,	82,	85],
            [11,	12,	35,	36,	59,	60,	83,	84]])
    else:
        print("Error. Only 72 and 96 cells modules supported at the moment. Change numcells to either of this options!")
        return
    
    if portraitorlandscape == 'landscape':
        stdpl = stdpl.transpose()
    elif portraitorlandscape != 'portrait':
        print("Error. portraitorlandscape variable must either be 'landscape' or 'portrait'")
        return
    
    cellsx = len(stdpl[1]); cellsy = len(stdpl)
    
    return stdpl, cellsx, cellsy


def calculatePVMismatch(pvsys, stdpl, cellsx, cellsy, Gpoat):
    r''' calls PVMismatch with all the pre-generated values on bifacial_radiance
    
    Example:
    PowerAveraged, PowerDetailed = def calculatePVMismatch(pvsys, stdpl, cellsx, cellsy, Gpoat)

    '''

    import numpy as np
    
    if np.mean(Gpoat) < 0.001:
        PowerAveraged = 0
        PowerDetailed = 0
    else:                                   
        # makes the system  # 1 module, in portrait mode.         
        G=np.array([Gpoat]).transpose()
        H = np.ones([1,cellsx]) 
        array_det = np.dot(G,H) 
        array_avg = np.ones([cellsy,cellsx])*np.mean(Gpoat)        
                                
        # ACtually do calculations
        pvsys.setSuns({0: {0: [array_avg, stdpl]}})
        PowerAveraged=pvsys.Pmp
        
        pvsys.setSuns({0: {0: [array_det, stdpl]}})
        PowerDetailed=pvsys.Pmp

    return PowerAveraged, PowerDetailed

def mad_fn(data):
    '''
    Mean average deviation calculation for mismatch purposes.
    
    Parameters
    ----------
    data : np.ndarray
        Gtotal irradiance measurements.

    Returns
    -------
    scalar :   return MAD / Average for a 1D array
    
    Equation: 1/(n^2*Gavg)*Sum Sum (abs(G_i - G_j))
    ## Note: starting with Pandas 1.0.0 this function will not work on Series objects.
    '''
    import numpy as np
    import pandas as pd
    # Pandas returns a notimplemented error if this is a series.
    if type(data) == pd.Series:
        data = data.to_numpy()
    
    return (np.abs(np.subtract.outer(data,data)).sum()/float(data.__len__())**2 / np.mean(data))*100



def analysisIrradianceandPowerMismatch(testfolder, writefiletitle, portraitorlandscape, bififactor, numcells=72, downsamplingmethod='byCenter'):
    r'''
    Use this when sensorsy calculated with bifacial_radiance > cellsy
    
    Reads and calculates power output and mismatch for each file in the 
    testfolder where all the bifacial_radiance irradiance results .csv are saved.
    First load each file, cleans it and resamples it to the numsensors set in this function,
    and then calculates irradiance mismatch and PVMismatch power output for averaged, minimum,
    or detailed irradiances on each cell for the cases of A) only 12 or 8 downsmaples values are
    considered (at the center of each cell), and B) 12 or 8 values are obtained from averaging
    all the irradiances falling in the area of the cell (No edges or inter-cell spacing are considered
    at this moment). Then it saves all the A and B irradiances, as well as the cleaned/resampled
    front and rear irradiances.
    
    Ideally sensorsy in the read data is >> 12 to give results for the irradiance mismatch in the cell.
    
    Also ideally n
     
    Parameters
    ----------
    testfolder:   folder containing output .csv files for bifacial_radiance
    writefiletitle:   .csv title where the output results will be saved. 
    portraitorlandscape: 'portrait' or 'landscape', for PVMismatch input
                      which defines the electrical interconnects inside the module. 
    bififactor: bifaciality factor of the module. Max 1.0. ALL Rear irradiance values saved include the bifi-factor.
    downsampling method: 1 - 'byCenter' - 2 - 'byAverage'
    
    Example:
        
    # User information.
    import bifacial_radiance
    testfolder=r'C:\Users\sayala\Documents\HPC_Scratch\EUPVSEC\HPC Tracking Results\RICHMOND\Bifacial_Radiance Results\PVPMC_0\results'
    writefiletitle= r'C:\Users\sayala\Documents\HPC_Scratch\EUPVSEC\HPC Tracking Results\RICHMOND\Bifacial_Radiance Results\PVPMC_0\test_df.csv'
    sensorsy=100
    portraitorlandscape = 'portrait'
    analysis.analysisIrradianceandPowerMismatch(testfolder, writefiletitle, portraitorlandscape, bififactor=1.0, numcells=72)

    '''
    from bifacial_radiance import load
    import os, glob
    import pandas as pd
        
    # Default variables 
    numpanels=1 # 1 at the moment, necessary for the cleaning routine.
    automatic=True
    
    #loadandclean
    # testfolder = r'C:\Users\sayala\Documents\HPC_Scratch\EUPVSEC\PinPV_Bifacial_Radiance_Runs\HPCResults\df4_FixedTilt\FixedTilt_Cairo_C_0.15\results'
    filelist = sorted(os.listdir(testfolder)) 
    #filelist = sorted(glob.glob(os.path.join('testfolder','*.csv'))) 
    print('{} files in the directory'.format(filelist.__len__()))

    # Check number of sensors on data.
    temp = load.read1Result(os.path.join(testfolder,filelist[0]))
    sensorsy = len(temp)

    # Setup PVMismatch parameters
    stdpl, cellsx, cellsy = _setupforPVMismatch(portraitorlandscape=portraitorlandscape, sensorsy=sensorsy, numcells=numcells)

    F=pd.DataFrame()
    B=pd.DataFrame()
    for z in range(0, filelist.__len__()):
            data=load.read1Result(os.path.join(testfolder,filelist[z]))
            [frontres, backres] = load.deepcleanResult(data, sensorsy=sensorsy, numpanels=numpanels, automatic=automatic)
            F[filelist[z]]=frontres
            B[filelist[z]]=backres          

    B = B*bififactor
    # Downsample routines:
    if sensorsy > cellsy:
        if downsamplingmethod == 'byCenter':
            print("Sensors y > cellsy; Downsampling data by finding CellCenter method")
            F = _sensorsdownsampletocellbyCenter(F, cellsy)
            B = _sensorsdownsampletocellbyCenter(B, cellsy)
        elif downsamplingmethod == 'byAverage':
            print("Sensors y > cellsy; Downsampling data by Averaging data into Cells method")
            F = _sensorsdownsampletocellsbyAverage(F, cellsy)
            B = _sensorsdownsampletocellsbyAverage(B, cellsy)
        else:
            print ("Sensors y > cellsy for your module. Select a proper downsampling method ('byCenter', or 'byAverage')")
            return
    elif sensorsy < cellsy:
        print("Sensors y < cellsy; Upsampling data by Interpolation")
        F = _sensorupsampletocellsbyInterpolation(F, cellsy)
        B = _sensorupsampletocellsbyInterpolation(B, cellsy)
    elif sensorsy == cellsy:
        print ("Same number of sensorsy and cellsy for your module.")
        F = F
        B = B

    # Calculate POATs
    Poat = F+B
            
    # Define arrays to fill in:
    Pavg_all=[]; Pdet_all=[]
    Pavg_front_all=[]; Pdet_front_all=[]
    colkeys = F.keys()
    
    import pvmismatch
    
    if cellsx*cellsy == 72:
        cell_pos = pvmismatch.pvmismatch_lib.pvmodule.STD72
    elif cellsx*cellsy == 96:
        cell_pos = pvmismatch.pvmismatch_lib.pvmodule.STD96
    else:
        print("Error. Only 72 and 96 cells modules supported at the moment. Change numcells to either of this options!")
        return
    
    pvmod=pvmismatch.pvmismatch_lib.pvmodule.PVmodule(cell_pos=cell_pos)        
    pvsys = pvmismatch.pvsystem.PVsystem(numberStrs=1, numberMods=1, pvmods=pvmod)  


    # Calculate powers for each hour:
    for i in range(0,len(colkeys)):        
        Pavg, Pdet = calculatePVMismatch(pvsys = pvsys, stdpl=stdpl, cellsx=cellsx, cellsy=cellsy, Gpoat=list(Poat[colkeys[i]]/1000))
        Pavg_front, Pdet_front = calculatePVMismatch(pvsys = pvsys, stdpl = stdpl, cellsx = cellsx, cellsy = cellsy, Gpoat= list(F[colkeys[i]]/1000))
        Pavg_all.append(Pavg)
        Pdet_all.append(Pdet)
        Pavg_front_all.append(Pavg_front) 
        Pdet_front_all.append(Pdet_front)
    
    ## Rename Rows and save dataframe and outputs.
    F.index='FrontIrradiance_cell_'+F.index.astype(str)
    B.index='BackIrradiance_cell_'+B.index.astype(str)
    Poat.index='POAT_Irradiance_cell_'+Poat.index.astype(str)
    
    ## Transpose 
    F = F.T
    B = B.T
    Poat = Poat.T

    # Statistics Calculatoins
    dfst=pd.DataFrame()
    dfst['MAD/G_Total'] = mad_fn(Poat.T)
    dfst['Front_MAD/G_Total'] = mad_fn(F.T)
    dfst['MAD/G_Total**2'] = dfst['MAD/G_Total']**2
    dfst['Front_MAD/G_Total**2'] = dfst['Front_MAD/G_Total']**2
    dfst['poat'] = Poat.mean(axis=1)
    dfst['gfront'] = F.mean(axis=1)
    dfst['grear'] = B.mean(axis=1)
    dfst['bifi_ratio'] =  dfst['grear']/dfst['gfront']
    dfst['stdev'] = Poat.std(axis=1)/ dfst['poat']
    dfst.index=Poat.index.astype(str)

    # Power Calculations/Saving
    Pout=pd.DataFrame()
    Pout['Pavg']=Pavg_all
    Pout['Pdet']=Pdet_all
    Pout['Front_Pavg']=Pavg_front_all
    Pout['Front_Pdet']=Pdet_front_all
    Pout['Mismatch_rel'] = 100-(Pout['Pdet']*100/Pout['Pavg'])
    Pout['Front_Mismatch_rel'] = 100-(Pout['Front_Pdet']*100/Pout['Front_Pavg'])   
    Pout.index=Poat.index.astype(str)

    ## Save CSV
    df_all = pd.concat([Pout,dfst,Poat,F,B],axis=1)
    df_all.to_csv(writefiletitle)
    print("Saved Results to ", writefiletitle)
    

    
        
